from injector import inject

from ..base import ConnectionTargetAdapter
from ..domain.enums import ConnectionTypes
from ..types.bigdata.adapters.target import BigDataTargetAdapter
from ..types.sql.adapters.target import SqlTargetAdapter
from ..types.webservice.adapters.target import WebServiceTargetAdapter
from ....dependency import IScoped
from ....exceptions import IncompatibleAdapterException, NotSupportedFeatureException


class ConnectionTargetAdapterFactory(IScoped):
    @inject
    def __init__(self,
                 sql_target_adapter: SqlTargetAdapter,
                 big_data_target_adapter: BigDataTargetAdapter,
                 web_service_target_adapter: WebServiceTargetAdapter,
                 # file_adapter: FileAdapter,
                 # queue_adapter: QueueAdapter,
                 ):
        # self.queue_adapter = queue_adapter
        # self.file_adapter = file_adapter
        self.web_service_target_adapter = web_service_target_adapter
        self.big_data_target_adapter = big_data_target_adapter
        self.sql_target_adapter = sql_target_adapter

    def get_adapter(self, connection_type: ConnectionTypes) -> ConnectionTargetAdapter:
        if connection_type == ConnectionTypes.Sql:
            if isinstance(self.sql_target_adapter, ConnectionTargetAdapter):
                return self.sql_target_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.sql_target_adapter} is incompatible with ConnectionTargetAdapter")
        elif connection_type == ConnectionTypes.File:
            if isinstance(self.file_target_adapter, ConnectionTargetAdapter):
                return self.file_target_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.file_target_adapter} is incompatible with ConnectionTargetAdapter")
        elif connection_type == ConnectionTypes.Queue:
            if isinstance(self.queue_target_adapter, ConnectionTargetAdapter):
                return self.queue_target_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.queue_target_adapter} is incompatible with ConnectionTargetAdapter")
        elif connection_type == ConnectionTypes.BigData:
            if isinstance(self.big_data_target_adapter, ConnectionTargetAdapter):
                return self.big_data_target_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.big_data_target_adapter} is incompatible with ConnectionTargetAdapter")
        elif connection_type == ConnectionTypes.WebService:
            if isinstance(self.web_service_target_adapter, ConnectionTargetAdapter):
                return self.web_service_target_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.web_service_target_adapter} is incompatible with ConnectionTargetAdapter")
        elif connection_type == ConnectionTypes.InMemory:
            if isinstance(self.in_memory_target_adapter, ConnectionTargetAdapter):
                return self.in_memory_target_adapter
            else:
                raise IncompatibleAdapterException(
                    f"{self.in_memory_target_adapter} is incompatible with ConnectionTargetAdapter")
        else:
            raise NotSupportedFeatureException(f"{connection_type.name}")



from injector import inject

from .operation_execution_initializer import OperationExecutionInitializer
from ....operation.domain import OperationBase
from .....dependency import IScoped


class DefaultOperationExecutionInitializer(OperationExecutionInitializer, IScoped):
    @inject
    def __init__(self):
        pass

    def initialize(self, operation: OperationBase) -> OperationBase:
        return operation



from injector import inject
from sqlalchemy.exc import OperationalError

from .seed import Seed
from ..base import DatabaseSessionManager
from ...dependency import IScoped
from ...dependency.provider import ServiceProvider
from ...logging.loggers.sql import SqlLogger


class SeedRunner(IScoped):
    @inject
    def __init__(self,
                 logger: SqlLogger,
                 service_provider: ServiceProvider,
                 ):
        self.service_provider = service_provider
        self.logger = logger

    def run(self):
        try:
            # trying to connect database first
            database_session_manager = self.service_provider.get(DatabaseSessionManager)
            database_session_manager.connect()
            for seed_class in Seed.__subclasses__():
                try:
                    instance = self.service_provider.get(seed_class)
                    instance.seed()
                except Exception as ex:
                    self.logger.exception(ex, "Class instance not found on container.")
                    instance = seed_class()
                    instance.seed()

        except OperationalError as opex:
            self.logger.exception(opex, "Database connection getting error on running seeds.")
        except Exception as ex:
            self.logger.exception(ex, "Seeds getting error.")
        finally:
            self.service_provider.get(DatabaseSessionManager).close()



from __future__ import absolute_import

from injector import inject
from sqlalchemy import create_engine, pool
from sqlalchemy.future import Engine
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy.pool import StaticPool

from ...configuration.models.database import DatabaseConfig
from ...dependency import IScoped
from ...utils.utils import Utils


class DatabaseSessionManager(IScoped):
    @inject
    def __init__(self,
                 database_config: DatabaseConfig):
        self.database_config = database_config
        self.engine: Engine = None
        self._SessionFactory: sessionmaker = None
        self.session: Session = None
        self.connect()

    def __del__(self):
        # close = getattr(self, "close", None)
        # if callable(close):
        #     self.close()
        pass

    def connect(self):
        if self.database_config.type is not None:
            if self.database_config.connection_string is None or self.database_config.connection_string == '':
                self.database_config.connection_string = Utils.get_connection_string(
                    database_config=self.database_config)
            if self.database_config.type == 'SQLITE':
                self.engine = create_engine(self.database_config.connection_string,
                                            connect_args={"check_same_thread": False},
                                            poolclass=StaticPool,
                                            execution_options=self.database_config.execution_options)
            else:
                self.engine = create_engine(self.database_config.connection_string,
                                            poolclass=pool.NullPool,
                                            pool_pre_ping=True,
                                            connect_args={"application_name": self.database_config.application_name})
            self._SessionFactory = sessionmaker(bind=self.engine)
            self.session: Session = self.session_factory()

    def dispose(self):
        if self.engine is not None:
            self.engine.dispose()

    def close(self):
        if self.session is not None:
            self.session.close()
        if self.engine is not None:
            self.dispose()
            # self.engine = None

    def session_factory(self):
        if self._SessionFactory is not None:
            self.session = self._SessionFactory()
        return self.session

    def commit(self):
        if self.session is not None:
            self.session.commit()

    def rollback(self):
        if self.session is not None:
            self.session.flush()
            self.session.rollback()



from ..base import ChannelQueue
from ..domain import TaskMessage


class Publisher:
    def __init__(self, channel: ChannelQueue):
        self.channel = channel

    def publish(self, message: TaskMessage):
        self.channel.put(message)



from abc import abstractmethod
from typing import List

from ...integration.domain.base import IntegrationBase


class ConnectionTargetAdapter:
    def clear_data(
            self,
            integration: IntegrationBase
    ) -> int:
        pass

    @abstractmethod
    def write_data(
            self,
            integration: IntegrationBase,
            source_data: List[any]
    ) -> int:
        pass

    @abstractmethod
    def do_target_operation(
            self,
            integration: IntegrationBase
    ) -> int:
        pass



from abc import abstractmethod
from datetime import datetime

from ....utils import Utils


class ILogger:

    def log_init(self):
        """
        initialization of log file.
        """
        pass

    def prepare_message(message):
        log_datetime = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        process_info = Utils.get_process_info()
        return f'{log_datetime} - {process_info} - {message} '

    @abstractmethod
    def log(self, level, message):
        pass

    @abstractmethod
    def exception(self, exception: Exception, message: str = None):
        pass

    def critical(self, message):
        pass

    @abstractmethod
    def fatal(self, message):
        pass

    @abstractmethod
    def error(self, message):
        pass

    @abstractmethod
    def warning(self, message):
        pass

    @abstractmethod
    def info(self, message):
        pass

    @abstractmethod
    def debug(self, message):
        pass



from dataclasses import dataclass
from typing import Optional

from pdip.integrator.connection.domain.authentication.basic import ConnectionBasicAuthentication
from pdip.integrator.connection.domain.enums import ConnectionTypes, ConnectorTypes
from pdip.integrator.connection.domain.server.base import ConnectionServer


@dataclass
class SqlConnectionConfiguration:
    Name: str = None
    ConnectionType: ConnectionTypes = None
    ConnectorType: ConnectorTypes = None
    ConnectionString: Optional[str] = None
    Driver: Optional[str] = None
    Server: ConnectionServer = None
    Sid: Optional[str] = None
    ServiceName: Optional[str] = None
    Database: Optional[str] = None
    BasicAuthentication: ConnectionBasicAuthentication = None
    ApplicationName: Optional[str] = None
    ExecutionOptions: Optional[str] = None



from dataclasses import dataclass

from pdip.integrator.connection.domain.enums import ConnectionTypes, ConnectorTypes


@dataclass
class InMemoryConnectionConfiguration:
    Name: str = None
    ConnectionString: str = None
    ConnectionType: ConnectionTypes = None
    ConnectorType: ConnectorTypes = None
    Database: str = None



from dataclasses import dataclass

from ..soap import SoapConfiguration
from .....authentication.basic import ConnectionBasicAuthentication
from .....enums import ConnectionTypes, ConnectorTypes
from .....server.base import ConnectionServer


@dataclass
class WebServiceConnectionConfiguration:
    Name: str = None
    ConnectionType: ConnectionTypes = None
    ConnectorType: ConnectorTypes = None
    Server: ConnectionServer = None
    Soap: SoapConfiguration = None
    BasicAuthentication: ConnectionBasicAuthentication = None
    Ssl: bool = False



import re
import time

from injector import inject

from .sql_connector import SqlConnector
from .sql_dialect import SqlDialect
from .sql_iterator import SqlIterator
from .sql_policy import SqlPolicy
from ......dependency import IScoped


def connect(func):
    def inner(*args, **kwargs):
        try:
            args[0].connector.connect()
            return func(*args, **kwargs)
        finally:
            args[0].connector.disconnect()

    return inner


class SqlContext(IScoped):
    @inject
    def __init__(self,
                 policy: SqlPolicy,
                 retry_count=3):
        self.connector: SqlConnector = policy.connector
        self.dialect: SqlDialect = policy.dialect
        self.retry_count = retry_count
        self.default_retry = 1

    @connect
    def get_query_columns(self, query, excluded_columns=None):
        self.connector.cursor.execute(query)
        if excluded_columns is not None:
            columns = [column[0] for column in self.connector.cursor.description if
                       column[0] not in excluded_columns]
        else:
            columns = [column[0] for column in self.connector.cursor.description]
        return columns

    @connect
    def fetch_query(self, query, excluded_columns=None):
        self.connector.cursor.execute(query)
        if excluded_columns is not None:
            columns = [column[0] for column in self.connector.cursor.description if column[0] not in excluded_columns]
        else:
            columns = [column[0] for column in self.connector.cursor.description]
        results = []
        for row in self.connector.cursor.fetchall():
            results.append(dict(zip(columns, row)))
        return results

    @connect
    def get_table_count(self, schema, table):
        count_query = self.dialect.get_table_count_query(schema=schema, table=table)
        self.connector.cursor.execute(count_query)
        datas = self.connector.cursor.fetchall()
        return datas[0][0]

    @connect
    def get_count_for_query(self, query):
        count_query = self.dialect.get_count_query(query=query)
        self.connector.cursor.execute(count_query)
        datas = self.connector.cursor.fetchall()
        return datas[0][0]

    def get_table_data(self, schema, table, columns=None):
        data_query = self.dialect.prepare_select_query(schema=schema, table=table, columns=columns)
        return self.fetch_query(data_query)

    def get_data_for_query(self, query):
        data_query = self.dialect.get_select_query(query=query)
        return self.fetch_query(data_query)

    def get_data_with_paging(self, query, start, end):
        data_query = self.dialect.get_paging_query(
            query=query,
            start=start,
            end=end
        )
        results = self.fetch_query(data_query, excluded_columns=['row_number'])
        return results

    def get_iterator_for_query(self, query, limit):
        data_query = self.dialect.get_select_query(query=query)
        iterator = SqlIterator(connector=self.connector, query=data_query, limit=limit)
        return iterator

    def get_iterator(self, schema, table, columns=None, limit=0):
        data_query = self.dialect.prepare_select_query(schema=schema, table=table, columns=columns)
        iterator = SqlIterator(connector=self.connector, query=data_query, limit=limit)
        return iterator

    @connect
    def execute(self, query) -> any:
        self.connector.cursor.execute(query)
        self.connector.connection.commit()
        return self.connector.cursor.rowcount

    @connect
    def execute_many(self, query, data):
        return self._execute_with_retry(query=query, data=data, retry=self.default_retry)

    def _execute_many_start(self, query, data):
        return self.connector.execute_many(query=query, data=data)

    def _execute_with_retry(self, query, data, retry):
        try:
            return self._execute_many_start(query=query, data=data)
        except Exception as ex:
            if retry > self.retry_count:
                print(f"Db write error on Error:{ex}")
                raise
            print(
                f"Getting error on insert (Operation will be retried. Retry Count:{retry}). Error:{ex}")
            # retrying connect to db,
            self.connector.connect()
            time.sleep(1)
            return self._execute_with_retry(query=query, data=data, retry=retry + 1)

    def create_table(self, schema, table, columns, if_exists=None):
        query = self.dialect.get_create_table_query(schema=schema, table=table, columns=columns, if_exists=if_exists)
        return self.execute(query=query)

    def drop_table(self, schema, table):
        query = self.dialect.get_drop_table_query(schema=schema, table=table)
        return self.execute(query=query)

    def truncate_table(self, schema, table):
        query = self.dialect.get_truncate_table_query(schema=schema, table=table)
        return self.execute(query=query)

    @staticmethod
    def replace_regex(text, field, indexer):
        text = re.sub(r'\(:' + field + r'\b', f'({indexer}', text)
        text = re.sub(r':' + field + r'\b\)', f'{indexer})', text)
        text = re.sub(r':' + field + r'\b', f'{indexer}', text)
        return text

    def prepare_target_query(self, column_rows, query):
        target_query = query
        for column_row in column_rows:
            index = column_rows.index(column_row)
            indexer = self.dialect.indexer.format(index=index)
            target_query = self.replace_regex(target_query, column_row, indexer)
        return target_query

    def prepare_insert_row(self, data, columns):
        insert_rows = []
        for extracted_data in data:
            row = []
            for column_row in columns:
                prepared_data = extracted_data[column_row]
                row.append(prepared_data)
            insert_rows.append(tuple(row))
        return insert_rows

    def _get_values_query(self, source_column_count):
        indexer_array = []
        for index in range(source_column_count):
            column_indexer = self.dialect.indexer.format(index=index)
            indexer_array.append(column_indexer)
        values_query = ','.join(indexer_array)
        return values_query

    def get_target_query(self, query, source_columns, target_columns, schema, table, source_column_count):
        if query is not None:
            if source_columns is not None and len(source_columns) > 0:
                column_rows = source_columns
                query = self.prepare_target_query(
                    column_rows=column_rows,
                    query=query
                )
            else:
                if schema is None or schema == '' or table is None or table == '':
                    raise Exception(f"Schema and table required. {schema}.{table}")
                values_query = self._get_values_query(source_column_count)
                query = self.dialect.get_insert_values_query(
                    values_query=values_query,
                    schema=schema,
                    table=table
                )
        else:
            if source_columns is not None and len(source_columns) > 0:
                if schema is None or schema == '' or table is None or table == '':
                    raise Exception(f"Schema and table required. {schema}.{table}")
                target_column_rows = [self.dialect.mark_to_object(column) for column in target_columns]
                columns_query = ','.join(target_column_rows)
                values_query = self._get_values_query(source_column_count)
                query = self.dialect.get_insert_query(
                    columns_query=columns_query,
                    values_query=values_query,
                    schema=schema,
                    table=table
                )
            else:
                if schema is None or schema == '' or table is None or table == '':
                    raise Exception(f"Schema and table required. {schema}.{table}")
                values_query = self._get_values_query(source_column_count)
                query = self.dialect.get_insert_values_query(
                    values_query=values_query,
                    schema=schema,
                    table=table
                )
        return query

    def get_insert_values(self, columns):
        insert_row_values = ''
        for column in columns:
            insert_row_values += f'{"" if column == columns[0] else ", "}:{column.Name}'
        return insert_row_values

    def get_insert_columns(self, columns):
        insert_row_columns = ''
        for column in columns:
            insert_row_columns += f'{"" if column == columns[0] else ", "}{self.dialect.mark_to_object(column.Name)}'
        return insert_row_columns

    def generate_insert_query(self, schema, table, source_columns, target_columns):
        insert_values = self.get_insert_values(source_columns)
        insert_columns = self.get_insert_columns(target_columns)
        query = self.dialect.get_insert_query(schema=schema, table=table, columns_query=insert_columns,
                                              values_query=insert_values)
        return query



from .default_integrator_initializer import DefaultIntegratorInitializer
from .integrator_initializer import IntegratorInitializer
from .integrator_initializer_factory import IntegratorInitializerFactory



import queue
import threading

from pdip.logging.loggers.console import ConsoleLogger
from .channel_queue import ChannelQueue
from ..domain import TaskMessage


class EventListener(threading.Thread):
    def __init__(self,
                 channel: ChannelQueue,
                 subscribers: {},
                 logger: ConsoleLogger,
                 *args, **kwargs
                 ):
        super().__init__(*args, **kwargs)
        self.logger = logger
        self.subscribers = subscribers
        self.channel = channel
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    def run(self):
        while True:
            try:
                task: TaskMessage = self.channel.get()
                if task is None:
                    return
                if task.event in self.subscribers.keys():
                    for callback in self.subscribers[task.event]:
                        callback(**task.kwargs)
                else:
                    self.logger.warning("Event {0} has no subscribers".format(task.event))
                if task.is_finished:
                    break
                self.channel.done()
            except queue.Empty:
                return
            except Exception as ex:
                self.logger.exception(ex, f'Event listener getting error.')
                return
            finally:
                pass



from injector import inject

from .in_memory_context import InMemoryContext
from .in_memory_policy import InMemoryPolicy
from ....domain.enums import ConnectorTypes, ConnectionTypes
from ....domain.types.inmemory import InMemoryConnectionConfiguration
from ......dependency import IScoped


class InMemoryProvider(IScoped):
    @inject
    def __init__(self):
        pass

    def __initialize_context(self, config: InMemoryConnectionConfiguration):
        policy = InMemoryPolicy(config=config)
        context = InMemoryContext(policy=policy)
        return context

    def get_context_by_config(self, config: InMemoryConnectionConfiguration) -> InMemoryContext:
        return self.__initialize_context(config=config)

    def get_context(
            self,
            connector_type: ConnectorTypes,
            database: str) -> InMemoryContext:
        """
        Creating Context
        """
        if connector_type == connector_type.SqLite:
            config = InMemoryConnectionConfiguration(
                ConnectionType=ConnectionTypes.InMemory,
                ConnectorType=ConnectorTypes.SqLite,
                Database=database
            )
        else:
            raise Exception(f"{connector_type.name} connector type not supported")

        return self.__initialize_context(config=config)



from dataclasses import dataclass
from typing import Optional

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class ConnectionColumnBase:
    Name: str = None
    Type: Optional[str] = None
    Nullable: Optional[str] = None



from injector import inject

from .default_operation_execution_initializer import DefaultOperationExecutionInitializer
from .operation_execution_initializer import OperationExecutionInitializer
from .....dependency import IScoped
from .....dependency.provider import ServiceProvider


class OperationExecutionInitializerFactory(IScoped):
    @inject
    def __init__(
            self,
            service_provider: ServiceProvider
    ):
        self.service_provider = service_provider

    def get(self) -> OperationExecutionInitializer:
        subclasses = OperationExecutionInitializer.__subclasses__()
        if subclasses is not None and len(subclasses) > 0:
            if len(subclasses) > 1:
                initializer_classes = [subclass for subclass in subclasses if
                                       subclass != DefaultOperationExecutionInitializer]
                initializer_class = initializer_classes[0]
            else:
                initializer_class = subclasses[0]
            initializer = self.service_provider.get(initializer_class)
            return initializer



import pyodbc

from ...base.big_data_connector import BigDataConnector
from .....domain.authentication.mechanism import MechanismTypes
from .....domain.types.bigdata.configuration.base import BigDataConnectionConfiguration


class ImpalaConnector(BigDataConnector):
    def __init__(self, config: BigDataConnectionConfiguration):
        self.config: BigDataConnectionConfiguration = config
        if self.config.ConnectionString is not None and self.config.ConnectionString != '' and not self.config.ConnectionString.isspace():
            self.connection_string = self.config.ConnectionString
        else:
            if self.config.Driver is None or self.config.Driver == '':
                self.config.Driver = self.find_driver_name()
            self.connection_string = f'Driver={self.config.Driver};Host={self.config.Server.Host};Port={self.config.Server.Port};'
            if self.config.Database is not None and self.config.Database != '':
                self.connection_string += f'Database={self.config.Database};'
            if self.config.AuthenticationMechanismType == MechanismTypes.NoAuthentication:
                pass
            elif self.config.AuthenticationMechanismType == MechanismTypes.Kerberos:
                self.connection_string += f'AuthMech=1;KrbRealm={self.config.KerberosAuthentication.KrbRealm};KrbFQDN={self.config.KerberosAuthentication.KrbFqdn};KrbServiceName={self.config.KerberosAuthentication.KrbServiceName};'
            elif self.config.AuthenticationMechanismType == MechanismTypes.UserName:
                self.connection_string += f'AuthMech=2;UID={self.config.BasicAuthentication.User};'
            elif self.config.AuthenticationMechanismType == MechanismTypes.UserNamePassword:
                self.connection_string += f'AuthMech=3;UID={self.config.BasicAuthentication.User};PWD={self.config.BasicAuthentication.Password};'
            if self.config.UseOnlySspi:
                self.connection_string += f'UseOnlySSPI=1;'
            if self.config.Ssl:
                self.connection_string += f'SSL=1;'

        self.connection = None
        self.cursor = None

    def connect(self):
        self.connection = pyodbc.connect(self.connection_string, autocommit=True)
        self.cursor = self.connection.cursor()

    def disconnect(self):
        try:
            if self.cursor is not None:
                self.cursor.close()

            if self.connection is not None:
                self.connection.close()
        except Exception:
            pass

    def find_driver_name(self):
        drivers = pyodbc.drivers()
        driver_name = None
        driver_names = [x for x in drivers if 'Impala' in x or 'impala' in x]
        if driver_names:
            driver_name = list(reversed(driver_names))[0]
        return driver_name

    def get_connection(self):
        return self.connection

    def execute_many(self, query, data):
        self.cursor.fast_executemany = True
        try:
            self.cursor.executemany(query, data)
            self.connection.commit()
            return self.cursor.rowcount
        except Exception as error:
            try:
                self.connection.rollback()
                self.cursor.fast_executemany = False
                self.cursor.executemany(query, data)
                self.connection.commit()
                return self.cursor.rowcount
            except Exception as error:
                self.connection.rollback()
                self.cursor.close()
                raise



from dataclasses import dataclass

from ..base.base_config import BaseConfig


@dataclass
class ApiConfig(BaseConfig):
    base_url: str = None
    doc_url: str = None
    is_debug: bool = None
    version: str = None
    port: int = None
    origins: str = None
    authorizations: any = None
    security: any = None



import json
from datetime import datetime
from typing import List

from injector import inject

from pdip.exceptions import NotSupportedFeatureException
from pdip.integrator.connection.base import ConnectionTargetAdapter
from pdip.integrator.connection.types.webservice.base import WebServiceProvider
from pdip.integrator.integration.domain.base import IntegrationBase
from pdip.json import date_time_parser


class WebServiceTargetAdapter(ConnectionTargetAdapter):
    @inject
    def __init__(self,
                 provider: WebServiceProvider,
                 ):
        self.provider = provider

    def clear_data(self, integration: IntegrationBase) -> int:
        raise NotSupportedFeatureException(f"{self.__class__.__name__} {self.clear_data.__name__}")

    def prepare_insert_row(self, data, columns):
        insert_rows = []
        for extracted_data in data:
            row = []
            for column in columns:
                column_data = extracted_data[column]
                row.append(column_data)

            insert_rows.append(tuple(row))
        return insert_rows

    def prepare_data(self, integration: IntegrationBase, source_data: any) -> List[any]:
        # columns = integration.SourceConnections.Columns
        # if columns is not None:
        #     source_columns = [(column.Name) for column in columns]
        # elif columns is None:
        #     source_columns = source_data[0].keys()
        # if isinstance(source_data, DataFrame):
        #     data = source_data[source_columns]
        #     prepared_data = data.values.tolist()
        # else:
        #     prepared_data = self.prepare_insert_row(data=source_data, columns=source_columns)
        # data = source_data[source_column_rows]
        # prepared_data = data.values.tolist()
        return source_data

    def prepare_target_query(self, integration: IntegrationBase, source_column_count: int) -> str:
        raise NotSupportedFeatureException(f"{self.__class__.__name__} {self.prepare_target_query.__name__}")

    def write_data(self, integration: IntegrationBase, prepared_data: List[any]) -> int:
        if prepared_data is not None and len(prepared_data) > 0:
            target_context = self.provider.get_context_by_config(
                config=integration.TargetConnections.WebService.Connection)
            affected_row_count = 0
            for d in prepared_data:
                request_data = integration.TargetConnections.WebService.RequestBody
                for key in d.keys():
                    request_data = target_context.replace_regex(request_data, key, f'{d[key]}')

                request_data = target_context.replace_regex(request_data, 'CurrentDate', f'{datetime.now()}')
                json_data = json.loads(request_data, object_hook=date_time_parser)
                request = target_context.prepare_request_auth(request=json_data)
                result = target_context.call_service(method=integration.TargetConnections.WebService.Method,
                                                     request=request)
                affected_row_count += 1
            return affected_row_count
        else:
            return 0

    def do_target_operation(self, integration: IntegrationBase) -> int:
        raise NotSupportedFeatureException(f"{self.__class__.__name__} {self.do_target_operation.__name__}")



from injector import inject

from .operation_integration_execution_initializer import OperationIntegrationExecutionInitializer
from ....operation.domain import OperationIntegrationBase
from .....dependency import IScoped


class DefaultOperationIntegrationExecutionInitializer(OperationIntegrationExecutionInitializer, IScoped):
    @inject
    def __init__(self):
        pass

    def initialize(self, operation_integration: OperationIntegrationBase) -> OperationIntegrationBase:
        return operation_integration



from typing import Generic, TypeVar

from .command_query_handler_base import CommandQueryHandlerBase
from .iquery import IQuery

QH = TypeVar('QH', covariant=True, bound=IQuery)


class IQueryHandler(Generic[QH], CommandQueryHandlerBase[QH]):
    def __init__(self) -> QH:
        pass

    def handle(self, query: QH) -> any:
        pass



from typing import Generic, TypeVar

from .command_query_base import CommandQueryBase

Q = TypeVar('Q', covariant=True)


class IQuery(Generic[Q], CommandQueryBase[Q]):
    def __init__(self) -> Q:
        pass



from .default_operation_integration_execution_initializer import DefaultOperationIntegrationExecutionInitializer
from .operation_integration_execution_initializer import OperationIntegrationExecutionInitializer
from .operation_integration_execution_initializer_factory import OperationIntegrationExecutionInitializerFactory



import os

import pyodbc
from sqlalchemy import create_engine
from sqlalchemy.engine import URL

from ...base import SqlConnector
from .....domain.types.sql.configuration.base import SqlConnectionConfiguration


class ClickHouseConnector(SqlConnector):
    def __init__(self, config: SqlConnectionConfiguration):
        self.config: SqlConnectionConfiguration = config
        if self.config.ConnectionString is not None and self.config.ConnectionString != '' and not self.config.ConnectionString.isspace():
            self.connection_string = self.config.ConnectionString
        else:
            if self.config.Driver is None or self.config.Driver == '':
                self.config.Driver = self.find_driver_name()
            app_name = os.getenv('CLICKHOUSE_APP_NAME', 'pdi')
            self.connection_string = 'DRIVER={%s};SERVER=%s;DATABASE=%s;UID=%s;PWD=%s;APP=%s;timeout=360' % (
                self.config.Driver, self.config.Server.Host, self.config.Database,
                self.config.BasicAuthentication.User, self.config.BasicAuthentication.Password, app_name)
        self.connection = None
        self.cursor = None

    def connect(self):
        self.connection = pyodbc.connect(self.connection_string)  # ,ansi=True)
        # self.connection.setencoding(encoding='utf-8')
        self.cursor = self.connection.cursor()
        self.cursor.setinputsizes([(pyodbc.SQL_WVARCHAR, 0, 0)])

    def disconnect(self):
        try:
            if self.cursor is not None:
                self.cursor.close()

            if self.connection is not None:
                self.connection.close()
        except Exception:
            pass

    def get_connection(self):
        return self.connection

    def get_engine_connection_url(self):
        connection_url = URL.create(
            "clickhouse+native",
            query={"odbc_connect": self.connection_string}
        )
        return connection_url

    def get_engine(self):
        connection_url = self.get_engine_connection_url()
        engine = create_engine(connection_url)
        return engine

    def find_driver_name(self):
        drivers = pyodbc.drivers()
        driver_name = None
        driver_names = [x for x in drivers if 'ClickHouse ODBC Driver' in x]
        if driver_names:
            driver_name = list(reversed(driver_names))[0]
        else:

            driver_names = [
                x for x in drivers if 'ClickHouse' in x]
            if driver_names:
                driver_name = list(reversed(driver_names))[0]
            else:
                driver_name = drivers[0]
        return driver_name

    def execute_many(self, query, data):
        self.cursor.fast_executemany = True
        try:
            self.cursor.executemany(query, data)
            self.connection.commit()
            return self.cursor.rowcount
        except Exception as error:
            try:
                self.connection.rollback()
                self.cursor.fast_executemany = False
                self.cursor.executemany(query, data)
                self.connection.commit()
                return self.cursor.rowcount
            except Exception as error:
                self.connection.rollback()
                self.cursor.close()
                raise



from zeep import Client

from ...base.web_service_connector import WebServiceConnector
from .....domain.types.webservice.configuration.base import WebServiceConnectionConfiguration


class SoapConnector(WebServiceConnector):
    def __init__(self, config: WebServiceConnectionConfiguration):
        self.config: WebServiceConnectionConfiguration = config
        if config.Server.Port is not None:
            url = f'{config.Server.Host}:{config.Server.Port}{config.Soap.Wsdl}'
        else:
            url = f'{config.Server.Host}{config.Soap.Wsdl}'

        if config.Ssl:
            url = 'https://' + url
        else:
            url = 'http://' + url
        self.wsdl_url = url

        self.client: Client = None

    def connect(self):
        self.client = Client(self.wsdl_url)

    def disconnect(self):
        try:
            if self.client is not None:
                del self.client

        except Exception:
            pass

    def execute_many(self, query, data):
        self.cursor.fast_executemany = True
        try:
            self.cursor.executemany(query, data)
            self.connection.commit()
            return self.cursor.rowcount
        except Exception as error:
            try:
                self.connection.rollback()
                self.cursor.fast_executemany = False
                self.cursor.executemany(query, data)
                self.connection.commit()
                return self.cursor.rowcount
            except Exception as error:
                self.connection.rollback()
                self.cursor.close()
                raise

    def get_target_query_indexer(self):
        indexer = '?'
        return indexer

    def prepare_data(self, data):
        # if data is not None and isinstance(data, str):
        #     data = data\
        #         .replace("ı", "i")\
        #         .replace("ş", "s")\
        #         .replace("ğ", "g")\
        #         .replace("İ", "I")\
        #         .replace("Ş","S")\
        #         .replace("Ğ", "G")
        return data



from abc import ABC, abstractmethod

from ...operation.domain import OperationBase, OperationIntegrationBase


class IntegratorEventManager(ABC):
    @abstractmethod
    def log(self, data: any, message: str, exception=None):
        pass

    @abstractmethod
    def initialize(self, data: OperationBase):
        pass

    @abstractmethod
    def start(self, data: OperationBase):
        pass

    @abstractmethod
    def finish(self, data: OperationBase, exception=None):
        pass

    @abstractmethod
    def integration_initialize(self, data: OperationIntegrationBase, message):
        pass

    @abstractmethod
    def integration_start(self, data: OperationIntegrationBase, message):
        pass

    @abstractmethod
    def integration_finish(self, data: OperationIntegrationBase, data_count, message, exception=None):
        pass

    @abstractmethod
    def integration_target_truncate(self, data: OperationIntegrationBase, row_count):
        pass

    @abstractmethod
    def integration_execute_source(self, data: OperationIntegrationBase, row_count):
        pass

    @abstractmethod
    def integration_execute_target(self, data: OperationIntegrationBase, row_count):
        pass



from abc import abstractmethod

from ......dependency import IScoped


class InMemoryConnector(IScoped):
    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def get_connection(self):
        pass

    @abstractmethod
    def execute_many(self, query, data):
        pass

    @abstractmethod
    def get_target_query_indexer(self):
        pass

    @abstractmethod
    def get_truncate_query(self, schema, table):
        count_query = f'TRUNCATE TABLE {schema}.{table}'
        return count_query

    @abstractmethod
    def get_table_count_query(self, query):
        count_query = f"SELECT COUNT(*) as COUNT  FROM ({query})  as count_table"
        return count_query

    @abstractmethod
    def get_table_select_query(self, selected_rows, schema, table):
        return f'SELECT {selected_rows} FROM {schema}.{table}'

    @abstractmethod
    def get_table_data_query(self, query):
        return f"SELECT * FROM ({query}) base_query"

    @abstractmethod
    def get_table_data_with_paging_query(self, query, start, end):
        return f"SELECT * FROM ({query}) ordered_query   order by null limit {end - start} offset {start}"



# import os
# from queue import Queue
# from typing import List
#
# import pandas as pd
# from injector import inject
# from pandas import DataFrame
#
# from ..base import ConnectionAdapter
# from pdip.execution.connection.types.file import FileProvider
# from ....exceptions import NotSupportedFeatureException
# from ....logging.loggers.database import SqlLogger
#
#
# class FileAdapter(ConnectionAdapter):
#     @inject
#     def __init__(self,
#                  logger: SqlLogger,
#                  file_provider: FileProvider,
#                  operation_cache_service: OperationCacheService,
#                  ):
#         self.operation_cache_service = operation_cache_service
#         self.logger = logger
#         self.file_provider = file_provider
#
#     def clear_data(self, data_integration_id) -> int:
#         target_connection = self.operation_cache_service.get_target_connection(
#             data_integration_id=data_integration_id)
#         target_context = self.file_provider.get_context(
#             connection=target_connection.Connection)
#         data_integration_columns = self.operation_cache_service.get_columns_by_integration_id(
#             data_integration_id=data_integration_id)
#         file_path = os.path.join(target_connection.File.Folder, target_connection.File.FileName)
#         if target_connection.File.Csv.HasHeader:
#             if target_connection.File.Csv.Header is not None and target_connection.File.Csv.Header != '':
#                 headers = target_connection.File.Csv.Header.split(target_connection.File.Csv.Separator)
#             else:
#                 headers = [(data_integration_column.TargetColumnName) for data_integration_column in
#                            data_integration_columns]
#             truncate_affected_rowcount = target_context.recreate_file(
#                 file=file_path, headers=headers,
#                 separator=target_connection.File.Csv.Separator)
#         else:
#             truncate_affected_rowcount = target_context.delete_file(
#                 file=file_path)
#
#         return truncate_affected_rowcount
#
#     def get_source_data_count(self, data_integration_id) -> int:
#         return -1
#
#     def start_source_data_operation(self,
#                                     data_integration_id: int,
#                                     data_operation_job_execution_integration_id: int,
#                                     limit: int,
#                                     process_count: int,
#                                     data_queue: Queue,
#                                     data_result_queue: Queue):
#
#         source_connection = self.operation_cache_service.get_source_connection(
#             data_integration_id=data_integration_id)
#
#         source_context = self.file_provider.get_context(connection=source_connection.Connection)
#
#         has_header = None
#         if source_connection.File.Csv.HasHeader:
#             has_header = 0
#         headers = None
#         separator = source_connection.File.Csv.Separator
#         if source_connection.File.Csv.Header is not None and source_connection.File.Csv.Header != '':
#             headers = source_connection.File.Csv.Header.split(separator)
#         if source_connection.File.FileName is not None and source_connection.File.FileName != '':
#             file_path = source_context.get_file_path(folder_name=source_connection.File.Folder,
#                                                      file_name=source_connection.File.FileName)
#             source_context.get_unpredicted_data(file=file_path,
#                                                 names=headers,
#                                                 header=has_header,
#                                                 separator=separator,
#                                                 limit=limit,
#                                                 process_count=process_count,
#                                                 data_queue=data_queue,
#                                                 result_queue=data_result_queue)
#         else:
#
#             csv_files = source_context.get_all_files(folder_name=source_connection.File.Folder, file_regex='(.*csv$)')
#             for csv_file in csv_files:
#                 self.logger.info(f"file read started. FilePath:{csv_file} ")
#                 source_context.get_unpredicted_data(file=csv_file,
#                                                     names=headers,
#                                                     header=has_header,
#                                                     separator=separator,
#                                                     limit=limit,
#                                                     process_count=process_count,
#                                                     data_queue=data_queue,
#                                                     result_queue=data_result_queue)
#
#     def get_source_data(self, data_integration_id: int, paging_modifier: PagingModifier) -> DataFrame:
#         source_connection = self.operation_cache_service.get_source_connection(
#             data_integration_id=data_integration_id)
#         source_context = self.file_provider.get_context(connection=source_connection.Connection)
#
#         data_integration_columns = self.operation_cache_service.get_columns_by_integration_id(
#             data_integration_id=data_integration_id)
#         has_header = None
#         if source_connection.File.Csv.HasHeader:
#             has_header = 0
#         headers = None
#         if source_connection.File.Csv.Header is not None and source_connection.File.Csv.Header != '':
#             headers = source_connection.File.Csv.Header.split(source_connection.File.Csv.Separator)
#
#         file_path = os.path.join(source_connection.File.Folder, source_connection.File.FileName)
#         readed_data = source_context.get_data(file=file_path,
#                                               names=headers,
#                                               header=has_header,
#                                               start=paging_modifier.Start,
#                                               limit=paging_modifier.Limit,
#                                               separator=source_connection.File.Csv.Separator)
#         column_names = [(data_integration_column.SourceColumnName) for data_integration_column in
#                         data_integration_columns]
#         data = readed_data[column_names]
#         replaced_data = data.where(pd.notnull(data), None)
#         return replaced_data.values.tolist()
#
#     def prepare_insert_row(self, data, columns):
#         insert_rows = []
#         for extracted_data in data:
#             row = []
#             for column in columns:
#                 column_data = extracted_data[column]
#                 row.append(column_data)
#             insert_rows.append(tuple(row))
#         return insert_rows
#
#     def prepare_data(self, data_integration_id: int, source_data: DataFrame) -> List[any]:
#
#         data_integration_columns = self.operation_cache_service.get_columns_by_integration_id(
#             data_integration_id=data_integration_id)
#
#         source_columns = [(data_integration_column.SourceColumnName) for data_integration_column in
#                           data_integration_columns]
#         if isinstance(source_data, pd.DataFrame):
#             data = source_data[source_columns]
#             prepared_data = data.values.tolist()
#         else:
#             prepared_data = self.prepare_insert_row(data=source_data, columns=source_columns)
#         return prepared_data
#
#     def write_target_data(self, data_integration_id: int, prepared_data: List[any], ) -> int:
#         target_connection = self.operation_cache_service.get_target_connection(
#             data_integration_id=data_integration_id)
#
#         target_context = self.file_provider.get_context(connection=target_connection.Connection)
#         df = pd.DataFrame(prepared_data)
#         data = df.where(pd.notnull(df), None)
#         file_path = os.path.join(target_connection.File.Folder, target_connection.File.FileName)
#         affected_row_count = target_context.write_to_file(file=file_path,
#                                                           data=data,
#                                                           separator=target_connection.File.Csv.Separator)
#         return affected_row_count
#
#     def do_target_operation(self, data_integration_id: int) -> int:
#         raise NotSupportedFeatureException("File Target Operation")



import importlib

from injector import inject

from .web_service_connector import WebServiceConnector
from ....domain.enums import ConnectorTypes
from ....domain.types.webservice.configuration.base import WebServiceConnectionConfiguration


class WebServicePolicy:
    @inject
    def __init__(self, config: WebServiceConnectionConfiguration):
        self.config = config
        self.connector: WebServiceConnector = None
        self.connector_name = None
        connector_base_module = "pdip.integrator.connection.types.webservice.connectors"
        if self.config.ConnectorType == ConnectorTypes.Impala:
            connector_namespace = "soap"
            connector_name = "SoapConnector"
        else:
            raise Exception("Connector type not found")
        module = importlib.import_module(".".join([connector_base_module, connector_namespace]))
        connector_class = getattr(module, connector_name)
        if connector_class is not None:
            self.connector: WebServiceConnector = connector_class(config)



import datetime
import json


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            encoded_object = obj.isoformat()
        else:
            encoded_object = json.JSONEncoder.default(self, obj)
        return encoded_object



from .web_service_connector import WebServiceConnector
from .web_service_context import WebServiceContext
from .web_service_policy import WebServicePolicy
from .web_service_provider import WebServiceProvider



from injector import inject

from pdip.dependency import IScoped
from pdip.integrator.connection.factories import ConnectionTargetAdapterFactory
from pdip.integrator.domain.enums.events import EVENT_EXECUTION_INTEGRATION_EXECUTE_TRUNCATE, \
    EVENT_EXECUTION_INTEGRATION_EXECUTE_TARGET
from pdip.integrator.integration.domain.base import IntegrationBase
from pdip.integrator.operation.domain.operation import OperationIntegrationBase
from pdip.integrator.pubsub.base import ChannelQueue
from pdip.integrator.pubsub.domain import TaskMessage
from pdip.integrator.pubsub.publisher import Publisher
from ...base import IntegrationAdapter


class TargetIntegration(IntegrationAdapter, IScoped):
    @inject
    def __init__(self,
                 connection_target_adapter_factory: ConnectionTargetAdapterFactory
                 ):
        self.connection_target_adapter_factory = connection_target_adapter_factory

    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ) -> int:
        publisher = Publisher(channel=channel)
        target_adapter = self.connection_target_adapter_factory.get_adapter(
            connection_type=operation_integration.Integration.TargetConnections.ConnectionType)
        if operation_integration.Integration.IsTargetTruncate:
            truncate_affected_row_count = target_adapter.clear_data(integration=operation_integration.Integration)
            publisher.publish(message=TaskMessage(event=EVENT_EXECUTION_INTEGRATION_EXECUTE_TRUNCATE,
                                                  kwargs={"data": operation_integration,
                                                          "row_count": truncate_affected_row_count
                                                          }))
        affected_row_count = target_adapter.do_target_operation(integration=operation_integration.Integration)
        publisher.publish(message=TaskMessage(event=EVENT_EXECUTION_INTEGRATION_EXECUTE_TARGET,
                                              kwargs={"data": operation_integration,
                                                      "row_count": affected_row_count
                                                      }))
        return affected_row_count

    def get_start_message(self, integration: IntegrationBase):
        return f"Target integration started."

    def get_finish_message(self, integration: IntegrationBase, data_count: int):
        return f"Target integration finished."

    def get_error_message(self, integration: IntegrationBase):
        return f"Target integration getting error."



import importlib
import os
from multiprocessing.process import current_process
from typing import TypeVar, Type

from injector import singleton, Injector, threadlocal, Binder

from ...configuration import ConfigManager
from ...configuration.models.api import ApiConfig
from ...configuration.models.application import ApplicationConfig
from ...configuration.models.database import DatabaseConfig
from ...dependency import ISingleton, IScoped
from ...logging.loggers.console.console_logger import ConsoleLogger
from ...utils import ModuleFinder
from ...utils import Utils

T = TypeVar('T')


class ServiceProvider:
    """
    """

    def __init__(self,
                 root_directory: str = None,
                 configurations: [] = None,
                 excluded_modules: [] = None):
        self.root_directory = root_directory
        self.excluded_modules = excluded_modules
        self.api_provider = None
        self.injector = Injector()
        self.config_manager: ConfigManager = None
        self.module_finder: ModuleFinder = None
        self.binder: Injector = None

        self.logger = ConsoleLogger()
        # self.logger = logging.getLogger('test')
        # process_info = Utils.get_process_info()
        # logging.basicConfig(
        #     level=logging.INFO, format=f"%(asctime)s:%(name)s:%(levelname)s: - {process_info} - %(message)s")
        self.logger.info(f"Application initialize started")
        if configurations is not None:
            self.modules = [self.configure] + configurations
        else:
            self.modules = [self.configure]

        self.configure_startup(self.root_directory)
        self.process_info()

    def __del__(self):
        if hasattr(self, 'api_provider') and self.api_provider is not None:
            del self.api_provider
        self.module_finder.cleanup()
        if hasattr(self, 'logger'):
            del self.logger

    def get(self, instance_type: Type[T]) -> T:
        return self.injector.get(instance_type)

    def configure_startup(self, root_directory):
        # Importing all modules for dependency
        self.module_finder = ModuleFinder(root_directory=root_directory)
        # excluded_modules = ['controllers']
        excluded_modules = []
        if self.excluded_modules is not None:
            excluded_modules += self.excluded_modules
        self.module_finder.import_modules(excluded_modules=excluded_modules)

        self.config_manager = ConfigManager(
            root_directory=root_directory, module_finder=self.module_finder)

        # Configuration initialize
        application_config = self.config_manager.get(ApplicationConfig)
        if application_config is None or application_config.root_directory is None or application_config.root_directory == '':
            self.config_manager.set(
                ApplicationConfig, "root_directory", self.root_directory)
        self.set_config_values_name()

    def set_config_values_name(self):
        application_config = self.config_manager.get(ApplicationConfig)
        database_config: DatabaseConfig = self.config_manager.get(
            DatabaseConfig)
        if database_config is not None and database_config.application_name is None:
            process_info = self.get_process_info()
            hostname = os.getenv('HOSTNAME', '')
            self.config_manager.set(ApplicationConfig, "hostname", hostname)
            self.config_manager.set(
                DatabaseConfig, "application_name", f"{application_config.name}-({process_info})")
            if database_config.connection_string is None or database_config.connection_string == '':
                self.config_manager.set(DatabaseConfig, "connection_string", Utils.get_connection_string(
                    database_config=database_config, root_directory=self.root_directory))

    def configure(self, binder: Binder):
        self.binder = binder
        self.binder.bind(
            interface=ServiceProvider,
            to=self
        )
        self.binder.bind(
            interface=ModuleFinder,
            to=self.module_finder
        )
        self.binder.bind(
            interface=ConfigManager,
            to=self.config_manager,
            scope=singleton
        )

        for config in self.config_manager.get_all_type_configs():
            self.binder.bind(
                interface=config.get("type"),
                to=config.get("instance"),
                scope=singleton,
            )

        for singletonScope in ISingleton.__subclasses__():
            self.binder.bind(
                interface=singletonScope,
                to=singletonScope,
                scope=singleton,
            )

        for scoped in IScoped.__subclasses__():
            self.binder.bind(
                interface=scoped,
                to=scoped,
                scope=threadlocal,
            )
        del self.logger

    def get_process_info(self):
        return f"{current_process().name} ({os.getpid()},{os.getppid()})"

    def process_info(self):
        application_config: ApplicationConfig = self.config_manager.get(ApplicationConfig)
        if application_config is not None:
            hostname = f'-{application_config.hostname}' if (
                    application_config.hostname is not None and application_config.hostname != '') else ''
            self.logger.info(f"Application : {application_config.name}{hostname}")

    def initialize_injection(self, initialize_flask):
        if self.is_flask_api() and initialize_flask:

            application_config: ApplicationConfig = self.config_manager.get(ApplicationConfig)
            api_config: ApiConfig = self.config_manager.get(ApiConfig)

            api_namespace = "pdip.dependency.provider.api"
            api_provider_name = "ApiProvider"

            module = importlib.import_module(api_namespace)
            api_provider_class = getattr(module, api_provider_name)
            if api_provider_class is not None:
                self.api_provider = api_provider_class(
                    modules=self.modules,
                    application_config=application_config,
                    api_config=api_config,
                    injector=self.injector
                )
                self.api_provider.initialize()
        else:
            for module in self.modules:
                self.injector.binder.install(module)

    def is_flask_api(self):
        api_config: ApiConfig = self.config_manager.get(ApiConfig)
        return api_config is not None and api_config.is_debug is not None



import multiprocessing
from multiprocessing.managers import SyncManager

from .channel_queue import ChannelQueue
from .event_listener import EventListener
from .message_broker_worker import MessageBrokerWorker


class MessageBroker:
    def __init__(self, logger):
        self.logger = logger
        self.manager: SyncManager = None
        self.publish_queue = None
        self.message_queue = None
        self.publish_channel: ChannelQueue = None
        self.message_channel: ChannelQueue = None
        self.worker: MessageBrokerWorker = None
        self.listener: EventListener = None
        self.subscribers = {}
        self.max_join_time = 120

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self.worker is not None and self.worker.is_alive() and not self.worker.stopped():
            self.worker.stop()
        if self.listener is not None and self.listener.is_alive() and not self.listener.stopped():
            self.listener.stop()
        if self.manager is not None:
            self.manager.shutdown()

    def initialize(self):
        self.manager = multiprocessing.Manager()
        self.publish_queue = self.manager.Queue()
        self.message_queue = self.manager.Queue()
        self.publish_channel = ChannelQueue(channel_queue=self.publish_queue)
        self.message_channel = ChannelQueue(channel_queue=self.message_queue)
        self.worker: MessageBrokerWorker = MessageBrokerWorker(
            logger=self.logger,
            publish_channel=self.publish_channel,
            message_channel=self.message_channel,
        )

    def start(self):
        self.worker.start()
        self.listener = EventListener(
            channel=self.message_channel,
            subscribers=self.subscribers,
            logger=self.logger
        )
        self.listener.start()

    def join(self):
        self.worker.join(self.max_join_time)
        self.listener.join(self.max_join_time)

    def get_publish_channel(self):
        return self.publish_channel

    def subscribe(self, event, callback):
        if not callable(callback):
            raise ValueError("callback must be callable")
        if event is None or event == "":
            raise ValueError("Event cant be empty")

        if event not in self.subscribers.keys():
            self.subscribers[event] = [callback]
        else:
            self.subscribers[event].append(callback)

    def unsubscribe(self, event, callback):
        if event is not None or event != "" \
                and event in self.subscribers.keys():
            self.subscribers[event] = list(
                filter(
                    lambda x: x is not callback,
                    self.subscribers[event]
                )
            )
        else:
            self.logger.warning("Cant unsubscribe function '{0}' from event '{1}' ".format(event, callback))



from functools import lru_cache

from injector import inject

from .config_parameter_base import ConfigParameterBase
from ...data.repository import RepositoryProvider
from ...dependency import IScoped
from ...exceptions import RequiredClassException


class ConfigService(IScoped):
    @inject
    def __init__(self,
                 repository_provider: RepositoryProvider
                 ):
        self.repository_provider = repository_provider
        config_subclasses = ConfigParameterBase.__subclasses__()
        if config_subclasses is None or len(config_subclasses) == 0:
            raise RequiredClassException(f'Requires {ConfigParameterBase.__name__} derived class')
        config_class = config_subclasses[0]
        self.config_reposiotry = repository_provider.get(config_class)

    @lru_cache()
    def get_config_by_name(self, name):
        parameter = self.config_reposiotry.first(Name=name)
        if parameter is not None:
            return parameter.Value
        else:
            return None



class ProcessTask:
    def __init__(self,
                 Data: any = None,
                 SubProcessId: int = None,
                 IsFinished: bool = None,
                 IsProcessed: bool = None,
                 State: int = None,
                 Result: any = None,
                 Message: str = None,
                 Exception: Exception = None,
                 Traceback: bool = None):
        self.Data: any = Data
        self.SubProcessId: int = SubProcessId
        self.IsFinished: bool = IsFinished
        self.IsProcessed: bool = IsProcessed
        self.State: int = State
        self.Result: any = Result
        self.Message: str = Message
        self.Exception: Exception = Exception
        self.Traceback: str = Traceback



from abc import ABC, abstractmethod
from queue import Queue

from pandas import DataFrame

from pdip.dependency import IScoped


class QueueConnector(ABC, IScoped):
    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def delete_topic(self, topic_name):
        pass

    @abstractmethod
    def create_topic(self, topic_name):
        pass

    @abstractmethod
    def write_data(self, topic_name: str, messages: DataFrame):
        pass

    @abstractmethod
    def create_admin_client(self):
        pass

    @abstractmethod
    def create_producer(self):
        pass

    @abstractmethod
    def create_consumer(self, topic_name: str, auto_offset_reset: str, enable_auto_commit: bool, group_id: str):
        pass

    @abstractmethod
    def get_data(self, limit: int) -> DataFrame:
        pass

    @abstractmethod
    def get_unpredicted_data(self, limit: int, process_count: int, data_queue: Queue, result_queue: Queue) -> DataFrame:
        pass



import os
import shutil

from injector import inject

from ..configuration.models.application import ApplicationConfig
from ..dependency import IScoped


class FolderManager(IScoped):
    @inject
    def __init__(self,
                 application_config: ApplicationConfig
                 ):
        self.application_config = application_config

    def create_folder_if_not_exist(self, folder):
        if not os.path.exists(folder):
            os.makedirs(folder)

    def copytree(self, src, dst, symlinks=False, ignore=None):
        for item in os.listdir(src):
            s = os.path.join(src, item)
            d = os.path.join(dst, item)
            if os.path.isdir(s):
                shutil.copytree(s, d, symlinks, ignore)
            else:
                shutil.copy2(s, d)

    def get_old_folder_path(self, folder):

        if os.path.exists(folder + '_old'):
            return self.get_old_folder_path(folder + '1')
        else:
            return folder + '_old'

    def start_copy(self, folder):
        path = os.path.join(self.application_config.root_directory, folder)

        if os.path.exists(path):
            path_old = self.get_old_folder_path(path)
            self.create_folder_if_not_exist(path_old)
            self.copytree(path, path_old)



from .default_integrator_event_manager import DefaultIntegratorEventManager
from .integrator_event_manager import IntegratorEventManager
from .integrator_event_manager_factory import IntegratorEventManagerFactory



from dataclasses import dataclass
from typing import Optional

from ..configuration.base import BigDataConnectionConfiguration


@dataclass
class ConnectionBigDataBase:
    Connection: BigDataConnectionConfiguration = None
    Schema: Optional[str] = None
    ObjectName: Optional[str] = None
    Query: Optional[str] = None



import json
from typing import List

from injector import inject
from sqlalchemy import text

from ..configuration.models.application import ApplicationConfig
from ..dependency import IScoped
from ..json import JsonConvert


@JsonConvert.register
class Pagination:

    def __init__(self,
                 Filter: str = None,
                 Page: int = None,
                 PageUrl: str = None,
                 Limit: int = None,
                 TotalPage: int = None,
                 TotalCount: int = None
                 ):
        self.Filter: str = Filter
        self.Page: int = Page
        self.PageUrl: str = PageUrl
        self.Limit: int = Limit
        self.TotalPage: int = TotalPage
        self.TotalCount: int = TotalCount


class HtmlTemplateService(IScoped):
    @inject
    def __init__(self,
                 application_config: ApplicationConfig

                 ):
        self.application_config: ApplicationConfig = application_config

    @property
    def default_css(self):
        pagination_css = '''

            .pagination {
              display: table;
            margin: 0 auto;
                padding: 20px;
            }

            .pagination a {
              color: black;
              float: left;
              padding: 8px 16px;
              text-decoration: none;
              transition: background-color .3s;
              border: 1px solid #ddd;
            }

            .pagination a.active {
              background-color: #4CAF50;
              color: white;
              border: 1px solid #4CAF50;
            }

            .pagination a:hover:not(.active) {background-color: #ddd;}
'''
        return '''
            .wrapper{
                margin: 0 auto;
                padding: 20px;
            }    
            .container600 {
                width: 300px;
                max-width: 100%;
            }

            @media all and (max-width: 600px) {
                .container600 {
                    width: 100% !important;
                }
            }

            .col49 {
                width: 49%;
            }

            .col2 {
                width: 2%;
            }

            .col50 {
                width: 50%;
            }

            @media all and (max-width: 599px) {
                .fluid {
                    width: 100% !important;
                }
                .reorder {
                    width: 100% !important;
                    margin: 0 auto 10px;
                }
                .ghost-column {
                    display:none;
                    height:0;
                    width:0;
                    overflow:hidden;
                    max-height:0;
                    max-width:0;
                }
            }
            .pdi-column{
                text-align: left;
                padding:4px; 
                font-family: Arial,sans-serif; 
                font-size: 12px; 
                line-height:10px;
            }
            .pdi-row{
                text-align: left;
                padding:4px; 
                font-family: Arial,sans-serif; 
                font-size: 10px; 
                line-height:10px;
            }
            .row-nowrap{
                white-space: nowrap;
            }
            table {
              border-collapse: collapse;
              width: 100%;
            }

            th, td {
              text-align: left;
              padding: 8px;
            }

            tr:nth-child(even) {background-color: #f2f2f2;}

ul.breadcrumb {
  padding: 10px 16px;
  list-style: none;
  background-color: #eee;
}
ul.breadcrumb li {
  display: inline;
  font-size: 18px;
}
ul.breadcrumb li+li:before {
  padding: 8px;
  color: black;
  content: "/\00";
}
ul.breadcrumb li a {
  color: #0275d8;
  text-decoration: none;
}
ul.breadcrumb li a:hover {
  color: #01447e;
  text-decoration: underline;
}
            ''' + pagination_css

    def mail_html_template(self, body, mail_css=None):
        css = mail_css if mail_css is not None else self.default_css
        template = f'''
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml">
            <head>
                <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
                <title></title>
                <style>{css}</style>
            </head>
            <body>

                {body}
            </body>
        </html>
        '''
        return template

    def get_nullable_dict_value(self, dict, key):
        if key in dict:
            return dict[key]
        return None

    def get_dict_value(self, dict, key):
        if key in dict and dict[key] is not None:
            return dict[key]
        return ''

    def prepare_table_data_dynamic(self, query, headers, prepare_row, sortable=None, pagination: Pagination = None):

        if sortable is not None:
            query = query.order_by(text(sortable))
        pagination_json = None
        if pagination is not None:
            total_count = query.count()
            if pagination.Limit is None or pagination.Limit < 1 or pagination.Limit > 200:
                pagination.Limit = 50
            total_page = int(total_count / pagination.Limit) + 1
            if pagination.Page is None or pagination.Page < 1 or total_page < pagination.Page:
                pagination.Page = 1
            if pagination.Limit:
                query = query.limit(pagination.Limit)
            if pagination.Page:
                offset = (pagination.Page - 1) * pagination.Limit
                if offset is None or offset <= 0:
                    offset = 0
                query = query.offset(offset)
            pagination_json = {'PageUrl': pagination.PageUrl, 'PageNumber': pagination.Page, 'Limit': pagination.Limit,
                               'Count': total_count, 'TotalPage': total_page, 'Filter': pagination.Filter}
        rows = []
        for data in query:
            row = prepare_row(data)
            rows.append(row)
        return {'columns': headers, 'rows': rows,
                'pagination': pagination_json}

    def render_table(self, source, width=None):
        columns: List[str] = self.get_nullable_dict_value(source, 'columns')
        rows: List[str] = self.get_nullable_dict_value(source, 'rows')
        pagination_json = self.get_nullable_dict_value(source, 'pagination')
        headers = ''
        headers = headers + f'<th scope="col"  class="pdi-column">#</th>'
        for column in columns:
            column_style = self.get_dict_value(column, 'style')
            column_class = self.get_dict_value(column, 'class')
            column_value = self.get_dict_value(column, 'value')
            headers = headers + f'<th scope="col" style="{column_style}" class="pdi-column {column_class}">{column_value}</th>'
        bodies = ''
        index = 0
        for row in rows:
            bodies = bodies + '<tr>'
            index = index + 1
            bodies = bodies + f'<td valign="top" class="pdi-row ">{index}</td>'
            for data in row['data']:
                row_style = self.get_dict_value(data, 'style')
                row_class = self.get_dict_value(data, 'class')
                row_value = self.get_dict_value(data, 'value')
                bodies = bodies + f'<td valign="top" style="{row_style}" class="pdi-row {row_class}">{row_value}</td>'
            bodies = bodies + '</tr>'
        table_width = width if width is not None else '100%'
        pagination_html = ''
        if pagination_json is not None:
            page_data = ""
            # JsonConvert.register(Pagination)
            pagination = JsonConvert.FromJSON(json.dumps(pagination_json))

            # TotalPage = self.get_nullable_dict_value(pagination, 'TotalPage')
            for page in range(1, pagination.TotalPage + 1):
                filter = f'{pagination.Filter}' if pagination.Filter is not None and pagination.Filter != '' else ''
                page_url = pagination.PageUrl.format(f'?PageNumber={page}&Limit={pagination.Limit}&Filter={filter}')
                if page == pagination.PageNumber:
                    page_data = f'{page_data}<a href="{page_url}" class="active">{page}</a>'

                else:
                    page_data = f'{page_data}<a href="{page_url}" >{page}</a>'

            pagination_html = f'''
            <div class="pagination">
              {page_data}
            </div>
            '''
        table = f'''
        <table width="{table_width}" cellpadding="0" cellspacing="0" style="min-width:100%;">
                <thead>
                  {headers}
                </thead>
                <tbody>
                  {bodies}
                </tbody>
            </table>
            {pagination_html}
            '''
        return table

    def render_html(self,
                    content,
                    ):

        body_content = f'''

                <div class="wrapper">
                    {content}
                </div>
                '''
        mail_body = self.mail_html_template(body_content)
        return mail_body



import re

from injector import inject

from .web_service_policy import WebServicePolicy
from ......dependency import IScoped


class WebServiceContext(IScoped):
    @inject
    def __init__(self,
                 policy: WebServicePolicy,
                 retry_count=3):
        self.connector: WebServicePolicy = policy.connector
        self.retry_count = retry_count
        self.default_retry = 1

    def connect(func):
        def inner(*args, **kwargs):
            try:
                args[0].connector.connect()
                return func(*args, **kwargs)
            finally:
                args[0].connector.disconnect()

        return inner

    @connect
    def prepare_request_auth(self, request):
        request['Header']['userId'] = self.connector.config.BasicAuthentication.User
        request['Header']['password'] = self.connector.config.BasicAuthentication.User
        return request

    @connect
    def call_service(self, method, request):
        method = getattr(self.connector.client.service, method)
        result = method(request)
        return result

    @staticmethod
    def replace_regex(text, field, indexer):
        text = re.sub(r'\(:' + field + r'\b', f'({indexer}', text)
        text = re.sub(r':' + field + r'\b\)', f'{indexer})', text)
        text = re.sub(r':' + field + r'\b', f'{indexer}', text)
        return text

    def prepare_request_body(self, column_rows, request_body):
        body = request_body
        for column_row in column_rows:
            index = column_rows.index(column_row)
            indexer = self.connector.get_target_query_indexer().format(index=index)
            body = self.replace_regex(body, column_row[0], indexer)
        return body



from injector import inject

from .integrator_initializer import IntegratorInitializer
from ..execution.base import ExecutionInitializerFactory
from ...domain.enums.events import EVENT_EXECUTION_INITIALIZED, EVENT_EXECUTION_FINISHED, \
    EVENT_EXECUTION_STARTED, EVENT_EXECUTION_INTEGRATION_INITIALIZED, EVENT_EXECUTION_INTEGRATION_STARTED, \
    EVENT_EXECUTION_INTEGRATION_FINISHED, EVENT_EXECUTION_INTEGRATION_EXECUTE_SOURCE, \
    EVENT_EXECUTION_INTEGRATION_EXECUTE_TARGET, EVENT_LOG, EVENT_EXECUTION_INTEGRATION_EXECUTE_TRUNCATE
from ...event.base import IntegratorEventManagerFactory
from ...operation.domain import OperationBase
from ...pubsub.base import MessageBroker
from ....dependency import IScoped


class DefaultIntegratorInitializer(IntegratorInitializer, IScoped):
    @inject
    def __init__(
            self,
            integrator_event_manager_factory: IntegratorEventManagerFactory = None,
            execution_initializer_factory: ExecutionInitializerFactory = None,
    ):
        self.execution_initializer_factory = execution_initializer_factory
        self.integrator_event_manager_factory = integrator_event_manager_factory

    def initialize(
            self,
            operation: OperationBase,
            message_broker: MessageBroker,
            execution_id: int = None,
            ap_scheduler_job_id: int = None
    ) -> OperationBase:
        message_broker.initialize()
        self.register_default_event_listeners(message_broker)
        message_broker.start()
        operation = self.execution_initializer_factory \
            .get() \
            .initialize(operation=operation,
                        execution_id=execution_id,
                        ap_scheduler_job_id=ap_scheduler_job_id
                        )
        return operation

    def register_default_event_listeners(
            self,
            message_broker: MessageBroker
    ):
        integrator_event_manager = self.integrator_event_manager_factory.get()
        message_broker.subscribe(EVENT_LOG, integrator_event_manager.log)
        message_broker.subscribe(EVENT_EXECUTION_INITIALIZED, integrator_event_manager.initialize)
        message_broker.subscribe(EVENT_EXECUTION_STARTED, integrator_event_manager.start)
        message_broker.subscribe(EVENT_EXECUTION_FINISHED, integrator_event_manager.finish)
        message_broker.subscribe(EVENT_EXECUTION_INTEGRATION_INITIALIZED,
                                 integrator_event_manager.integration_initialize)
        message_broker.subscribe(EVENT_EXECUTION_INTEGRATION_STARTED, integrator_event_manager.integration_start)
        message_broker.subscribe(EVENT_EXECUTION_INTEGRATION_FINISHED, integrator_event_manager.integration_finish)
        message_broker.subscribe(EVENT_EXECUTION_INTEGRATION_EXECUTE_TRUNCATE,
                                 integrator_event_manager.integration_target_truncate)
        message_broker.subscribe(EVENT_EXECUTION_INTEGRATION_EXECUTE_SOURCE,
                                 integrator_event_manager.integration_execute_source)
        message_broker.subscribe(EVENT_EXECUTION_INTEGRATION_EXECUTE_TARGET,
                                 integrator_event_manager.integration_execute_target)



from injector import inject

from ...domain.base import IntegrationBase
from ...types.base import IntegrationAdapter
from ...types.source.base import SourceIntegration
from ...types.sourcetotarget.base import SourceToTargetIntegration
from ...types.target.base import TargetIntegration
from .....dependency import IScoped
from .....exceptions import IncompatibleAdapterException


class IntegrationAdapterFactory(IScoped):
    @inject
    def __init__(self,
                 source_integration: SourceIntegration,
                 target_integration: TargetIntegration,
                 source_to_target_integration: SourceToTargetIntegration,
                 ):
        self.source_to_target_integration = source_to_target_integration
        self.source_integration = source_integration
        self.target_integration = target_integration

    def get(self, integration: IntegrationBase) -> IntegrationAdapter:
        if integration.TargetConnections is None or integration.TargetConnections.ConnectionName is None:
            raise Exception(
                f"Target connection required for integration")
        elif integration.SourceConnections is None or integration.SourceConnections.ConnectionName is None:
            if isinstance(self.target_integration, IntegrationAdapter):
                return self.target_integration
            else:
                raise IncompatibleAdapterException(
                    f"{self.target_integration} is incompatible with {IntegrationAdapter}")
        else:
            if isinstance(self.source_integration, IntegrationAdapter):
                return self.source_to_target_integration
            else:
                raise IncompatibleAdapterException(
                    f"{self.source_integration} is incompatible with {IntegrationAdapter}")



import multiprocessing


class ProcessInfo:
    def __init__(self,
                 Process: any = None,
                 SubProcessId: int = None,
                 IsFinished: bool = False):
        self.Process: multiprocessing.Process = Process
        self.SubProcessId: int = SubProcessId
        self.IsFinished: bool = IsFinished



from abc import abstractmethod

from ......dependency import IScoped


class BigDataConnector(IScoped):
    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def get_connection(self):
        pass

    @abstractmethod
    def execute_many(self, query, data):
        pass

    @abstractmethod
    def get_target_query_indexer(self):
        pass



import queue
import threading

from pdip.logging.loggers.console import ConsoleLogger
from .channel_queue import ChannelQueue
from ..domain import TaskMessage


class MessageBrokerWorker(threading.Thread):
    def __init__(self,
                 logger: ConsoleLogger,
                 publish_channel: ChannelQueue,
                 message_channel: ChannelQueue,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logger
        self.message_channel = message_channel
        self.publish_channel = publish_channel
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    def run(self):
        while True:
            try:
                work: TaskMessage = self.publish_channel.get()
                if work is None:
                    return
                self.message_channel.put(work)
                if work.is_finished:
                    break
                self.publish_channel.done()
            except queue.Empty:
                return
            except Exception as e:
                self.logger.error('error')
                return
            finally:
                pass



import inspect
import os
from typing import TypeVar, Type

from ..configuration.models.application import ApplicationConfig
from ..dependency.container import DependencyContainer

T = TypeVar('T')


class Pdi(object):
    def __init__(self,
                 root_directory: str = None,
                 configurations: [] = None,
                 excluded_modules: [] = None,
                 initialize_flask: bool = True
                 ):
        if root_directory is None:
            stack = inspect.stack()
            root_directory = self.get_root_directory(stack)
        self.excluded_modules = excluded_modules
        self.configurations = configurations
        self.root_directory = root_directory
        if DependencyContainer.Instance is None:
            DependencyContainer.initialize_service(root_directory=root_directory,
                                                   configurations=configurations,
                                                   excluded_modules=excluded_modules,
                                                   initialize_flask=initialize_flask)

    def cleanup(self):
        if DependencyContainer is not None:
            DependencyContainer.cleanup()

    def get_root_directory(self, stack):
        file_path = stack[1].filename
        directory = os.path.dirname(os.path.realpath(file_path))
        return directory

    def get(self, instance_type: Type[T]) -> T:
        return DependencyContainer.Instance.get(instance_type)

    def set_secret_key(self, key):
        DependencyContainer.Instance.config_manager.set(
            ApplicationConfig, 'secret_key', key)



import json

from pdip.json import DateTimeEncoder, BaseConverter


class JsonConvert(object):
    mappings = {}

    @classmethod
    def class_mapper(clsself, d):
        dataFrozenset = frozenset(tuple([attr for attr, val in d.items()]))
        for keys, cls in clsself.mappings.items():
            if keys.issuperset(d.keys()) and dataFrozenset.issuperset(keys):  # are all required arguments present?
                return cls(**d)
        else:
            # Raise exception instead of silently returning None
            raise ValueError(f'Unable to find a matching class for object: {d}')

    @classmethod
    def register(clsself, cls):
        clsself.mappings[frozenset(tuple([attr for attr, val in cls().__dict__.items()]))] = cls
        return cls

    @classmethod
    def ToJSON(clsself, obj):
        return json.dumps(dict(obj), cls=DateTimeEncoder, indent=4)

    @classmethod
    def FromJSON(clsself, json_str):
        return json.loads(json_str, object_hook=clsself.class_mapper)



# from queue import Queue
# from typing import List
#
# import pandas as pd
# from injector import inject
# from pandas import DataFrame
#
# from ..base import ConnectionAdapter
# from pdip.execution.connection.types.queue.base import QueueProvider
# from ....exceptions import NotSupportedFeatureException
#
#
# class QueueAdapter(ConnectionAdapter):
#     @inject
#     def __init__(self,
#                  queue_provider: QueueProvider,
#                  operation_cache_service: OperationCacheService,
#                  ):
#         self.operation_cache_service = operation_cache_service
#         self.queue_provider = queue_provider
#
#     def clear_data(self, data_integration_id) -> int:
#         target_connection = self.operation_cache_service.get_target_connection(
#             data_integration_id=data_integration_id)
#         target_context = self.queue_provider.get_context(connection=target_connection.Connection)
#         target_context.delete_topic(topic_name=target_connection.Queue.TopicName)
#         target_context.create_topic(topic_name=target_connection.Queue.TopicName)
#         return 1
#
#     def start_source_data_operation(self,
#                                     data_integration_id: int,
#                                     data_operation_job_execution_integration_id: int,
#                                     limit: int,
#                                     process_count: int,
#                                     data_queue: Queue,
#                                     data_result_queue: Queue):
#         source_connection = self.operation_cache_service.get_source_connection(
#             data_integration_id=data_integration_id)
#
#         source_context = self.queue_provider.get_context(connection=source_connection.Connection)
#         source_context.get_unpredicted_data(topic_name=source_connection.Queue.TopicName, group_id="test",
#                                             limit=limit,
#                                             process_count=process_count,
#                                             data_queue=data_queue,
#                                             result_queue=data_result_queue)
#
#     def get_source_data_count(self, data_integration_id) -> int:
#         return -1
#
#     def get_source_data(self, data_integration_id: int, paging_modifier: PagingModifier) -> DataFrame:
#         raise NotSupportedFeatureException("Queue Get Source Data")
#
#     def prepare_data(self, data_integration_id: int, source_data: List[any]) -> List[any]:
#         data_integration_columns = self.operation_cache_service.get_columns_by_integration_id(
#             data_integration_id=data_integration_id)
#
#         source_columns = [(data_integration_column.SourceColumnName) for data_integration_column in
#                           data_integration_columns]
#         if isinstance(source_data, pd.DataFrame):
#             data = source_data[source_columns]
#             prepared_data = data.values.tolist()
#         else:
#             prepared_data = self.prepare_insert_row(data=source_data, columns=source_columns)
#         return prepared_data
#
#     def write_target_data(self, data_integration_id: int, prepared_data: List[any], ) -> int:
#         target_connection = self.operation_cache_service.get_target_connection(
#             data_integration_id=data_integration_id)
#
#         target_context = self.queue_provider.get_context(connection=target_connection.Connection)
#         data_integration_columns = self.data_integration_column_service.get_columns_by_integration_id(
#             data_integration_id=data_integration_id)
#         columns = [(data_integration_column.TargetColumnName) for data_integration_column in
#                    data_integration_columns]
#         df = pd.DataFrame(prepared_data, columns=columns)
#         data = df.where(pd.notnull(df), None)
#         affected_row_count = target_context.write_data(
#             topic_name=target_connection.Queue.TopicName, messages=data)
#         return affected_row_count
#
#     def do_target_operation(self, data_integration_id: int) -> int:
#         raise NotSupportedFeatureException("Queue Target Operation")



import importlib

from injector import inject

from .sql_connector import SqlConnector
from .sql_dialect import SqlDialect
from ....domain.enums import ConnectorTypes
from ....domain.types.sql.configuration.base import SqlConnectionConfiguration


class SqlPolicy:
    @inject
    def __init__(self, config: SqlConnectionConfiguration):
        self.config = config
        self.connector: SqlConnector = None
        self.connector_name = None
        connector_base_module = "pdip.integrator.connection.types.sql.connectors"
        dialect_base_module = "pdip.integrator.connection.types.sql.dialects"
        if self.config.ConnectorType == ConnectorTypes.MSSQL:
            namespace = "mssql"
            connector_name = "MssqlConnector"
            dialect_name = "MssqlDialect"
        elif self.config.ConnectorType == ConnectorTypes.ORACLE:
            namespace = "oracle"
            connector_name = "OracleConnector"
            dialect_name = "OracleDialect"
        elif self.config.ConnectorType == ConnectorTypes.POSTGRESQL:
            namespace = "postgresql"
            connector_name = "PostgresqlConnector"
            dialect_name = "PostgresqlDialect"
        elif self.config.ConnectorType == ConnectorTypes.MYSQL:
            namespace = "mysql"
            connector_name = "MysqlConnector"
            dialect_name = "MysqlDialect"
        elif self.config.ConnectorType == ConnectorTypes.CLICKHOUSE:
            namespace = "clickhouse"
            connector_name = "ClickHouseConnector"
            dialect_name = "ClickHouseDialect"
        else:
            raise Exception("Connector type not found")
        module = importlib.import_module(".".join([connector_base_module, namespace]))
        connector_class = getattr(module, connector_name)
        if connector_class is not None:
            self.connector: SqlConnector = connector_class(self.config)
        module = importlib.import_module(".".join([dialect_base_module, namespace]))
        dialect_class = getattr(module, dialect_name)
        if dialect_class is not None:
            self.dialect: SqlDialect = dialect_class(self.connector)



from .domain.dao_generate_config import DaoGenerateConfig
from .domain.generator import Generator
from .domain.query_generate_config import QueryGenerateConfig
from ...configuration.models.application import ApplicationConfig
from ...io.file_manager import FileManager
from ...io.folder_manager import FolderManager
from ...utils import ModuleFinder


class QueryGenerator(Generator):
    def __init__(self,
                 application_config: ApplicationConfig,
                 module_finder: ModuleFinder,
                 folder_manager: FolderManager,
                 file_manager: FileManager):
        self.file_manager = file_manager
        self.folder_manager = folder_manager
        self.application_config = application_config
        self.module_finder = module_finder

    def generate(self, generate_config: QueryGenerateConfig):

        query_folder_path = "/".join([generate_config.base_directory, generate_config.domain, generate_config.name])

        self.folder_manager.start_copy(query_folder_path)

        self.__create_request_file(query_name=generate_config.name, query_folder_path=query_folder_path,
                                   has_paging=generate_config.has_paging)
        self.__create_dto_file(query_name=generate_config.name, query_folder_path=query_folder_path)
        self.__create_response_file(query_name=generate_config.name, query_folder_path=query_folder_path,
                                    is_list=generate_config.is_list, has_paging=generate_config.has_paging)
        self.__create_query_file(query_name=generate_config.name, query_folder_path=query_folder_path)
        self.__create_specifications_file(query_name=generate_config.name, query_folder_path=query_folder_path,
                                          dao=generate_config.dao, is_list=generate_config.is_list,
                                          has_paging=generate_config.has_paging)

        self.__create_mapping_file(query_name=generate_config.name, query_folder_path=query_folder_path,
                                   dao=generate_config.dao)
        self.__create_query_handler_file(query_name=generate_config.name, query_folder_path=query_folder_path,
                                         dao=generate_config.dao, is_list=generate_config.is_list,
                                         has_paging=generate_config.has_paging)

    def __create_query_file(self, query_name: str, query_folder_path: str):
        file_name = f"{query_name}Query"
        base_namespace = query_folder_path.replace('/', '.')
        request_namespace = f'from {base_namespace}.{query_name}Request import {query_name}Request'
        response_namespace = f'from {base_namespace}.{query_name}Response import {query_name}Response'
        request_attr = f'request: {query_name}Request = None'
        content = \
            f'''from dataclasses import dataclass
from infrastructure.cqrs.IQuery import IQuery
{request_namespace}
{response_namespace}


@dataclass
class {query_name}Query(IQuery[{query_name}Response]):
    {request_attr}
        '''
        self.file_manager.create_file(folder=query_folder_path, file_name=file_name, content=content)

    def __create_request_file(self, query_name: str, query_folder_path: str,
                              has_paging: bool = True, ):
        file_name = f"{query_name}Request"
        paging_parent = "(PagingParameter, OrderByParameter)" if has_paging else ""
        content = \
            f'''
from domain.common.decorators.requestclass import requestclass


@requestclass
class {query_name}Request{paging_parent}:
    # TODO:Request attributes
    pass
    '''
        self.file_manager.create_file(folder=query_folder_path, file_name=file_name,
                                      content=content)

    def __create_dto_file(self, query_name: str, query_folder_path: str):
        file_name = f"{query_name}Dto"

        content = \
            f'''
from domain.common.decorators.dtoclass import dtoclass


@dtoclass
class {query_name}Dto:
    # TODO:Dto attributes
    pass
    '''
        self.file_manager.create_file(folder=query_folder_path, file_name=file_name,
                                      content=content)

    def __create_response_file(self, query_name: str, query_folder_path: str, is_list: bool, has_paging: bool):
        query_request_file_name = f"{query_name}Response"

        base_namespace = query_folder_path.replace('/', '.')
        dto_namespace = f'from {base_namespace}.{query_name}Dto import {query_name}Dto'
        if is_list:
            if has_paging:
                attributes = \
                    f'''    Data: List[{query_name}Dto] = None
    PageNumber: int = None
    PageSize: int = None
    Count: int = None'''
            else:
                attributes = f"\tData: List[{query_name}Dto] = None"
        else:
            attributes = f"\tData: {query_name}Dto = None"
        content = \
            f'''
from typing import List
from domain.common.decorators.responseclass import responseclass
{dto_namespace}


@responseclass
class {query_name}Response:
{attributes}
    '''
        self.file_manager.create_file(folder=query_folder_path, file_name=query_request_file_name,
                                      content=content)

    def __create_specifications_file(self, query_name: str, query_folder_path: str, dao: DaoGenerateConfig,
                                     is_list: bool = True,
                                     has_paging: bool = True):
        file_name = f"{query_name}Specifications"

        base_namespace = query_folder_path.replace('/', '.')
        query_namespace = f'from {base_namespace}.{query_name}Query import {query_name}Query'
        orderable = {}
        pageable = {}
        if is_list:
            if has_paging:
                orderable[
                    'Namespace'] = f'from domain.common.specifications.OrderBySpecification import OrderBySpecification\n'
                pageable[
                    'Namespace'] = f'from domain.common.specifications.PagingSpecification import PagingSpecification\n'
        content = ''
        content += 'from injector import inject\n'
        content += 'from sqlalchemy.orm import Query\n'
        content += 'from infrastructure.dependency import IScoped\n'
        content += 'from infrastructure.data.RepositoryProvider import RepositoryProvider\n'
        content += f'{dao.namespace}\n'

        content += f'{query_namespace}\n'
        if has_paging:
            content += orderable['Namespace']
            content += '\n'
            content += pageable['Namespace']
            content += '\n'
        content += '\n'
        content += \
            f'''
class {query_name}Specifications(IScoped):
    @inject
    def __init__(self,\n'''
        content += f'''                 repository_provider: RepositoryProvider,\n'''
        if has_paging:
            content += f'''                 order_by_specification: OrderBySpecification,\n'''
            content += f'''                 paging_specification: PagingSpecification,\n'''
        content += f'''                 ):\n'''

        content += f'''        self.repository_provider = repository_provider\n'''
        if has_paging:
            content += f'''        self.paging_specification = paging_specification\n'''
            content += f'''        self.order_by_specification = order_by_specification\n'''
        else:
            content += f'''        \n'''
        content += \
            f'''
    def __specified_query(self, query: {query_name}Query) -> Query:
        repository = self.repository_provider.get({dao.name})
        specified_query = repository.table 
        # TODO:specify query
        return specified_query
        
    def specify(self, query: {query_name}Query) -> Query:
    
        data_query = self.__specified_query(query=query)\n'''
        if has_paging:
            content += f'''
        order_by = self.order_by_specification.specify(order_by_parameter=query.request)
        if order_by is not None:
            data_query = data_query.order_by(order_by)

        page_size, offset = self.paging_specification.specify(paging_parameter=query.request)
        if page_size is not None:
            data_query = data_query.limit(page_size)
        if offset is not None:
            data_query = data_query.offset(offset)\n'''
        content += '''        return data_query\n'''

        content += f'''
    def count(self, query: {query_name}Query) -> Query:
        return self.__specified_query(query=query).count()
    '''
        self.file_manager.create_file(folder=query_folder_path, file_name=file_name,
                                      content=content)

    def __create_mapping_file(self, query_name: str, query_folder_path: str, dao: DaoGenerateConfig):
        file_name = f"{query_name}Mapping"
        base_namespace = query_folder_path.replace('/', '.')
        dto_namespace = f'from {base_namespace}.{query_name}Dto import {query_name}Dto'
        content = \
            f'''
from typing import List
{dto_namespace}
{dao.namespace}


class {query_name}Mapping:
    @classmethod
    def to_dto(cls, entity: {dao.name}) -> {query_name}Dto:
        dto = {query_name}Dto()
        return dto

    @classmethod
    def to_dtos(cls, entities: List[{dao.name}]) -> List[{query_name}Dto]:
        result: List[{query_name}Dto] = []
        for entity in entities:
            dto = cls.to_dto(entity=entity)
            result.append(dto)
        return result'''
        self.file_manager.create_file(folder=query_folder_path, file_name=file_name,
                                      content=content)

    def __create_query_handler_file(self, query_name: str, query_folder_path: str, dao: DaoGenerateConfig,
                                    is_list: bool = True, has_paging: bool = True):
        file_name = f"{query_name}QueryHandler"
        base_namespace = query_folder_path.replace('/', '.')
        query_namespace = f'from {base_namespace}.{query_name}Query import {query_name}Query'
        mapping_namespace = f'from {base_namespace}.{query_name}Mapping import {query_name}Mapping'
        request_namespace = f'from {base_namespace}.{query_name}Request import {query_name}Request'
        response_namespace = f'from {base_namespace}.{query_name}Response import {query_name}Response'
        specifications_namespace = f'from {base_namespace}.{query_name}Specifications import {query_name}Specifications'
        content = \
            f'''
from injector import inject
{mapping_namespace}
{query_namespace}
{response_namespace}
{specifications_namespace}
from infrastructure.cqrs.IQueryHandler import IQueryHandler
from infrastructure.dependency import IScoped


class {query_name}QueryHandler(IQueryHandler[{query_name}Query], IScoped):
    @inject
    def __init__(self,
                 specifications: {query_name}Specifications):
        self.specifications = specifications

    def handle(self, query: {query_name}Query) -> {query_name}Response:
        result = {query_name}Response()\n'''
        if has_paging:
            content += f'''
        result.Count = self.specifications.count(query=query)

        result.PageNumber = query.request.PageNumber
        result.PageSize = query.request.PageSize\n'''

        content += \
            f'''        data_query = self.specifications.specify(query=query)\n'''
        if is_list:
            content += \
                f'''        result.Data = {query_name}Mapping.to_dtos(data_query)\n'''
        else:
            content += \
                f'''        result.Data = {query_name}Mapping.to_dto(data_query)\n'''
        content += \
            f'''        return result\n'''
        self.file_manager.create_file(folder=query_folder_path, file_name=file_name,
                                      content=content)



import importlib
from typing import Type, TypeVar

from injector import inject

from .command_query_base import CommandQueryBase
from .icommand import ICommand
from .icommand_handler import ICommandHandler
from .iquery import IQuery
from .iquery_handler import IQueryHandler
from ..dependency import IScoped
from ..dependency.provider import ServiceProvider

T = TypeVar('T', covariant=True)


class Dispatcher(IScoped):
    @inject
    def __init__(self, service_provider: ServiceProvider):
        self.service_provider = service_provider

    def find_handler(self, cq, handler_type: Type[T]) -> T:
        try:
            importlib.import_module(cq.__module__ + 'Handler')
        except Exception as ex:
            pass
        subclasses = handler_type.__subclasses__()
        for handler_class in subclasses:
            result = handler_type[cq.__class__] == handler_class.__orig_bases__[0]
            if result:
                instance = self.service_provider.get(handler_class)
                return instance

    def dispatch(self, cq: CommandQueryBase[T]) -> T:
        if isinstance(cq, IQuery):
            handler_type = IQueryHandler
        elif isinstance(cq, ICommand):
            handler_type = ICommandHandler
        else:
            raise Exception("Command or query not found")
        handler = self.find_handler(cq, handler_type)
        if handler is None:
            raise Exception("Handler not founded")
        result = handler.handle(cq)
        if isinstance(cq, IQuery):
            return result



from injector import inject

from ..base import IntegrationSourceToTargetExecuteStrategy
from ..parallelthread.base import ParallelThreadIntegrationExecute
from ..singleprocess.base import SingleProcessIntegrationExecute
from .......dependency import IScoped
from .......exceptions import IncompatibleAdapterException


class IntegrationSourceToTargetExecuteStrategyFactory(IScoped):
    @inject
    def __init__(self,
                 parallel_thread_integration_execute: ParallelThreadIntegrationExecute,
                 single_process_integration_execute: SingleProcessIntegrationExecute,
                 ):
        self.single_process_integration_execute = single_process_integration_execute
        self.parallel_thread_integration_execute = parallel_thread_integration_execute

    def get(self, process_count: int) -> IntegrationSourceToTargetExecuteStrategy:
        if process_count is not None and process_count > 1:
            if isinstance(self.parallel_thread_integration_execute, IntegrationSourceToTargetExecuteStrategy):
                return self.parallel_thread_integration_execute
            else:
                raise IncompatibleAdapterException(
                    f"{self.parallel_thread_integration_execute} is incompatible with {IntegrationSourceToTargetExecuteStrategy}")
        else:
            if isinstance(self.single_process_integration_execute, IntegrationSourceToTargetExecuteStrategy):
                return self.single_process_integration_execute
            else:
                raise IncompatibleAdapterException(
                    f"{self.single_process_integration_execute} is incompatible with {IntegrationSourceToTargetExecuteStrategy}")



from abc import ABC, abstractmethod

from ...domain.base import IntegrationBase
from ....operation.domain.operation import OperationIntegrationBase
from ....pubsub.base import ChannelQueue


class IntegrationAdapter(ABC):
    @abstractmethod
    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ):
        pass

    @abstractmethod
    def get_start_message(self, integration: IntegrationBase):
        pass

    @abstractmethod
    def get_finish_message(self, integration: IntegrationBase, data_count: int):
        pass

    @abstractmethod
    def get_error_message(self, integration: IntegrationBase):
        pass



from .sql_connector import SqlConnector
from .sql_context import SqlContext
from .sql_dialect import SqlDialect
from .sql_iterator import SqlIterator
from .sql_policy import SqlPolicy
from .sql_provider import SqlProvider



from os import getenv
from pathlib import Path

from setuptools import setup, find_packages

here = Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

env_version = getenv('PYPI_PACKAGE_VERSION', default='0.6.10')
version = env_version.replace('v', '')
setup(
    name='pdip',
    version=f'{version}',
    description='Python Data Integrator infrastructures package',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ahmetcagriakca/pdip',
    download_url=f'https://github.com/ahmetcagriakca/pdip/archive/refs/tags/v{version}.tar.gz',
    author='Ahmet Çağrı AKCA',
    author_email='ahmetcagriakca@gmail.com',
    license='MIT',
    packages=find_packages(
        exclude=["tests", "tests*", "test_*", '__pycache__', '*.__pycache__', '__pycache.*', '*.__pycache__.*']),
    zip_safe=False,
    keywords=['PDI', 'API', 'ETL', 'PROCESS', 'MULTIPROCESS', 'IO', 'CQRS', 'MSSQL', 'ORACLE', 'POSTGRES', 'MYSQL',
              'CSV'],
    python_requires='>=3.8',
    install_requires=[
        "dataclasses",
        "injector",
        "PyYAML",
        "SQLAlchemy"
    ],
    extras_require={
        "api": [
            "Flask==3.0.3",
            "Flask_Cors==4.0.0",
            "Flask-Ext==0.1",
            "Flask-Injector==0.15.0",
            "flask-restx==1.3.0",
            "markupsafe==2.1.5",
            "Werkzeug==3.0.2"
        ],
        "cryptography": [
            "cryptography==42.0.5",
            "Fernet==1.0.1"
        ],
        "integrator": [
            "cx_Oracle==8.3.0",
            "dataclasses-json==0.5.6",
            "func-timeout==4.3.5",
            "kafka-python==2.0.2",
            "mysql-connector-python==8.4.0",
            "pandas==2.2.2",
            "psycopg2-binary==2.9.9",
            "pyodbc==5.1.0"
        ],
        "preferred": [
            "dataclasses==0.6",
            "injector==0.21.0",
            "PyYAML==6.0.1",
            "SQLAlchemy==2.0.29"
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/ahmetcagriakca/pdip/issues',
        'Source': 'https://github.com/ahmetcagriakca/pdip',
    },
)



from dataclasses import dataclass


@dataclass
class KerberosAuthentication:
    Principal: str = None
    Password: str = None
    KrbRealm: str = None
    KrbFqdn: str = None
    KrbServiceName: str = None



import sys
from typing import TypeVar

from flask import Flask
from flask_injector import FlaskInjector, request
from flask_restx import Api
from injector import Injector, Binder, singleton
from werkzeug.utils import redirect

from pdip.logging.loggers.console.console_logger import ConsoleLogger
from ....api.base import ResourceBase
from ....api.base.controller_base import Controller
from ....configuration.models.api import ApiConfig
from ....configuration.models.application import ApplicationConfig

T = TypeVar('T')


class ApiProvider:
    def __init__(self,
                 modules: [] = None,
                 application_config: ApplicationConfig = None,
                 api_config: ApiConfig = None,
                 injector: Injector = None):
        self.modules = modules
        self.injector = injector
        self.api_config = api_config
        self.application_config = application_config
        self.app: Flask = None
        self.api: Api = None
        self.binder: Binder = None
        self.logger = ConsoleLogger()

    def __del__(self):
        del self.api
        del self.app

        if hasattr(self, 'logger'):
            del self.logger

    def initialize(self):
        self.initialize_flask()

    def api_configure(self, binder: Binder):
        self.binder = binder

        if self.app is not None:
            self.binder.bind(
                interface=Flask,
                to=self.app,
                scope=singleton
            )
        if self.api is not None:
            self.binder.bind(
                interface=Api,
                to=self.api,
                scope=singleton
            )
        for controller in ResourceBase.__subclasses__():
            self.binder.bind(
                interface=controller,
                to=controller,
                scope=request
            )
        del self.logger

    def initialize_flask(self):
        application_name = ''
        if self.application_config is not None and self.application_config.name is not None:
            application_name = self.application_config.name
        if self.api_config is not None:
            self.app = Flask(application_name)
            with self.app.app_context():
                base_url = '/'
                if self.api_config.base_url is not None:
                    base_url = self.api_config.base_url
                doc_url = '/documentation'
                if self.api_config.doc_url is not None:
                    doc_url = self.api_config.doc_url

                @self.app.route(base_url)
                def home_redirect():
                    return redirect(doc_url, code=302, Response=None)

                self.api = Api(self.app,
                               title=f'{application_name}',
                               version=self.api_config.version,
                               doc=doc_url,
                               base_url=base_url)
                self.find_resources()
                FlaskInjector(app=self.app, modules=[
                                                        self.api_configure] + self.modules, injector=self.injector)

    def find_resources(self):
        for resource in ResourceBase.__subclasses__():
            if resource.__module__ in sys.modules and sys.modules[resource.__module__].__file__.lower().startswith(
                    self.application_config.root_directory.lower()):
                namespace = None
                namespace_name = None
                route = None
                if hasattr(resource, 'namespace'):
                    namespace = resource.namespace
                if hasattr(resource, 'namespace_name'):
                    namespace_name = resource.namespace_name
                if hasattr(resource, 'route'):
                    route = resource.route

                con = Controller(
                    cls=resource,
                    api=self.api,
                    application_config=self.application_config,
                    namespace=namespace,
                    namespace_name=namespace_name,
                    route=route)
                con.create_route()



from dataclasses import dataclass

from ..base.base_config import BaseConfig


@dataclass
class ApsConfig(BaseConfig):
    coalesce: bool = None
    max_instances: str = None
    thread_pool_executer_count: bool = None
    process_pool_executer_count: int = None
    default_misfire_grace_time_date_job: int = None
    default_misfire_grace_time_cron_job: int = None



import multiprocessing
import traceback
from queue import Queue
from time import time

import pandas as pd
from func_timeout import FunctionTimedOut, func_set_timeout
from injector import inject
from pandas import DataFrame, notnull

from ...base import IntegrationSourceToTargetExecuteStrategy
from ......domain.base import IntegrationBase
from .......connection.domain.task import DataQueueTask
from .......connection.factories import ConnectionSourceAdapterFactory, ConnectionTargetAdapterFactory
from .......domain.enums.events import EVENT_LOG
from .......operation.domain import OperationIntegrationBase
from .......pubsub.base import ChannelQueue
from .......pubsub.domain import TaskMessage
from .......pubsub.publisher import Publisher
from ........data.decorators import transactionhandler
from ........dependency import IScoped
from ........dependency.container import DependencyContainer
from ........processing import ProcessManager
from ........processing.factories import ProcessManagerFactory


class ParallelIntegrationExecute(IntegrationSourceToTargetExecuteStrategy, IScoped):
    @inject
    def __init__(
            self,
            process_manager_factory: ProcessManagerFactory,
            connection_source_adapter_factory: ConnectionSourceAdapterFactory,
            connection_target_adapter_factory: ConnectionTargetAdapterFactory
    ):
        self.connection_source_adapter_factory = connection_source_adapter_factory
        self.connection_target_adapter_factory = connection_target_adapter_factory
        self.process_manager_factory = process_manager_factory

    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ) -> int:
        publisher = Publisher(channel=channel)
        try:
            try:
                manager = multiprocessing.Manager()
                source_data_process_manager = self.process_manager_factory.get()
                execute_data_process_manager = self.process_manager_factory.get()
                data_queue = manager.Queue()
                data_result_queue = manager.Queue()

                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"Source data process will start"
                        }
                    )
                )
                self.start_source_data_subprocess(
                    source_data_process_manager=source_data_process_manager,
                    channel=channel,
                    operation_integration=operation_integration,
                    data_queue=data_queue,
                    data_result_queue=data_result_queue
                )

                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"Source data process started"
                        }
                    )
                )

                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"Execute data process will start"
                        }
                    )
                )
                total_row_count = self.start_execute_data_subprocess(
                    execute_data_process_manager=execute_data_process_manager,
                    channel=channel,
                    operation_integration=operation_integration,
                    data_queue=data_queue,
                    data_result_queue=data_result_queue
                )

                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"Execute data process finished"
                        }
                    )
                )
            finally:
                manager.shutdown()
                del source_data_process_manager
                del execute_data_process_manager

            return total_row_count
        except Exception as ex:
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"Integration getting error.",
                        'exception': ex
                    }
                )
            )
            raise

    def start_source_data_subprocess(
            self,
            source_data_process_manager: ProcessManager,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ):
        source_data_kwargs = {
            "operation_integration": operation_integration,
            "channel": channel,
            "data_queue": data_queue,
            "data_result_queue": data_result_queue,
        }

        source_data_process_manager.start_processes(
            process_count=1,
            target_method=self.start_source_data_process,
            kwargs=source_data_kwargs
        )

    def start_execute_data_subprocess(
            self,
            execute_data_process_manager: ProcessManager,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ) -> int:
        total_row_count = 0
        execute_data_kwargs = {
            "channel": channel,
            "operation_integration": operation_integration,
            "data_queue": data_queue,
            "data_result_queue": data_result_queue,
        }
        execute_data_process_manager.start_processes(
            process_count=operation_integration.ProcessCount,
            target_method=self.start_execute_data_process,
            kwargs=execute_data_kwargs
        )
        execute_data_process_results = execute_data_process_manager.get_results()
        for result in execute_data_process_results:
            if result.Exception is not None:
                raise result.Exception
            if result.Result is not None:
                total_row_count = total_row_count + result.Result
        return total_row_count

    @staticmethod
    def start_source_data_process(
            sub_process_id,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ):
        return DependencyContainer.Instance.get(ParallelIntegrationExecute).start_source_data_operation(
            sub_process_id=sub_process_id,
            channel=channel,
            operation_integration=operation_integration,
            data_queue=data_queue,
            data_result_queue=data_result_queue,
        )

    @transactionhandler
    def start_source_data_operation(
            self,
            sub_process_id,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ):

        publisher = Publisher(channel=channel)
        publisher.publish(
            message=TaskMessage(
                event=EVENT_LOG,
                kwargs={
                    'data': operation_integration,
                    'message': f"Source data operation started on process. SubProcessId: {sub_process_id}"
                }
            )
        )
        try:

            source_adapter = self.connection_source_adapter_factory.get_adapter(
                connection_type=operation_integration.Integration.SourceConnections.ConnectionType
            )

            data_count = source_adapter.get_source_data_count(integration=operation_integration.Integration)
            if data_count > 0:
                transmitted_data_count = 0
                limit = operation_integration.Limit
                end = limit
                start = 0
                id = 0
                while True:
                    if end != limit and end - data_count >= limit:
                        break
                    id = id + 1
                    data_queue_task = DataQueueTask(
                        Id=id,
                        Data=None,
                        IsDataFrame=False,
                        Start=start,
                        End=end,
                        Limit=limit,
                        IsFinished=False
                    )
                    data_queue.put(data_queue_task)
                    transmitted_data_count = transmitted_data_count + 1
                    if transmitted_data_count >= operation_integration.ProcessCount:
                        result = data_result_queue.get()
                        if result:
                            transmitted_data_count = transmitted_data_count - 1
                        else:
                            break
                    end += limit
                    start += limit
            for i in range(operation_integration.ProcessCount):
                data_queue_finish_task = DataQueueTask(IsFinished=True)
                data_queue.put(data_queue_finish_task)

            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"Source data operation finished successfully. SubProcessId: {sub_process_id}"
                    }
                )
            )
        except Exception as ex:
            exception_traceback = traceback.format_exc()
            for i in range(operation_integration.ProcessCount):
                data_queue_error_task = DataQueueTask(
                    IsFinished=True,
                    Traceback=exception_traceback,
                    Exception=ex
                )
                data_queue.put(data_queue_error_task)

            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"Source data operation finished with error. SubProcessId: {sub_process_id}.",
                        'exception': ex

                    }
                )
            )
            raise

    @staticmethod
    def start_execute_data_process(
            sub_process_id,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ) -> int:
        return DependencyContainer.Instance.get(ParallelIntegrationExecute).start_execute_data_operation(
            sub_process_id=sub_process_id,
            channel=channel,
            operation_integration=operation_integration,
            data_queue=data_queue,
            data_result_queue=data_result_queue,
        )

    def start_execute_data_operation(
            self,
            sub_process_id: int,
            channel: ChannelQueue,
            operation_integration: OperationIntegrationBase,
            data_queue: Queue,
            data_result_queue: Queue
    ) -> int:

        publisher = Publisher(channel=channel)
        total_row_count = 0
        try:
            while True:
                data_task: DataQueueTask = data_queue.get()
                if data_task.IsFinished:
                    if data_task.Exception is not None:
                        exc = Exception(data_task.Traceback + '\n' + str(data_task.Exception))
                        raise exc
                    publisher.publish(
                        message=TaskMessage(
                            event=EVENT_LOG,
                            kwargs={
                                'data': operation_integration,
                                'message': f"{sub_process_id} process tasks finished"
                            }
                        )
                    )
                    return total_row_count
                else:
                    start = time()
                    data = data_task.Data
                    if data_task.IsDataFrame and data is not None:
                        source_data_json = data_task.Data
                        data: DataFrame = DataFrame(source_data_json)
                    data_count = 0
                    if data is None:
                        publisher.publish(
                            message=TaskMessage(
                                event=EVENT_LOG,
                                kwargs={
                                    'data': operation_integration,
                                    'message': f"{sub_process_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process got a new task without data"
                                }
                            )
                        )
                        data_count = self.start_execute_integration_with_paging(
                            integration=operation_integration.Integration,
                            start=data_task.Start,
                            end=data_task.End
                        )
                    elif data is not None and len(data) > 0:
                        if data_task.IsDataFrame and data_task.DataTypes is not None:
                            source_data = data.astype(dtype=data_task.DataTypes)
                        else:
                            source_data = data
                        if data_task.IsDataFrame:
                            source_data = source_data.where(notnull(data), None)
                            source_data = source_data.replace({pd.NaT: None})

                        publisher.publish(
                            message=TaskMessage(
                                event=EVENT_LOG,
                                kwargs={
                                    'data': operation_integration,
                                    'message': f"{sub_process_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process got a new task"
                                }
                            )
                        )
                        data_count = self.start_execute_integration_with_source_data(
                            integration=operation_integration.Integration,
                            source_data=source_data
                        )
                    else:
                        publisher.publish(
                            message=TaskMessage(
                                event=EVENT_LOG,
                                kwargs={
                                    'data': operation_integration,
                                    'message': f"{sub_process_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process got an empty task"
                                }
                            )
                        )
                    total_row_count = total_row_count + data_count
                    end = time()
                    publisher.publish(
                        message=TaskMessage(
                            event=EVENT_LOG,
                            kwargs={
                                'data': operation_integration,
                                'message': f"{sub_process_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process finished task. time:{end - start}"
                            }
                        )
                    )
                    data_task.IsProcessed = True
                    data_result_queue.put(True)
                data_queue.task_done()
        except FunctionTimedOut as fto:
            data_result_queue.put(False)
            end = time()
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"{sub_process_id}-{data_task.Message}:{data_task.Id}-{data_task.Start}-{data_task.End} process getting error. time:{end - start} - Error {fto.getMsg()}"
                    }
                )
            )
            raise Exception(f'Execution Operation timed out after {fto.timedOutAfter} seconds.')
        except Exception as ex:
            data_result_queue.put(False)
            raise

    @func_set_timeout(1800)
    def start_execute_integration_with_source_data(self,
                                                   integration: IntegrationBase,
                                                   source_data: any
                                                   ):
        target_adapter = self.connection_target_adapter_factory.get_adapter(
            connection_type=integration.TargetConnections.ConnectionType
        )
        target_adapter.write_data(
            integration=integration,
            source_data=source_data
        )
        return len(source_data)

    @func_set_timeout(1800)
    def start_execute_integration_with_paging(self,
                                              integration: IntegrationBase,
                                              start,
                                              end
                                              ):
        source_adapter = self.connection_source_adapter_factory.get_adapter(
            connection_type=integration.SourceConnections.ConnectionType
        )
        target_adapter = self.connection_target_adapter_factory.get_adapter(
            connection_type=integration.TargetConnections.ConnectionType
        )
        source_data = source_adapter.get_source_data_with_paging(
            integration=integration,
            start=start,
            end=end
        )
        target_adapter.write_data(
            integration=integration,
            source_data=source_data
        )
        return len(source_data)



import multiprocessing
from multiprocessing.managers import SyncManager
from queue import Queue
from typing import List

from .parallel_worker import ParallelWorker
from ..domain import ThreadInfo
from ..domain import ThreadTask
from ........domain.enums.events import EVENT_LOG
from ........pubsub.base import ChannelQueue
from ........pubsub.domain import TaskMessage
from ........pubsub.publisher import Publisher


class ParallelSessionBroker:
    def __init__(
            self,
            channel: ChannelQueue,
            operation_integration
    ):
        self.operation_integration = operation_integration
        self.publisher = Publisher(channel=channel)
        self._process_queue: Queue = None
        self._process_result_queue: Queue = None
        self.process_result_channel: ChannelQueue = None
        self.process_channel: ChannelQueue = None

        self._threads: List[ThreadInfo] = None
        self._thread_tasks: List[ThreadTask] = None
        self.channel = channel
        self.number_of_cores = multiprocessing.cpu_count()
        self._manager: SyncManager = None

    def __del__(self):
        self.__finish_all_threads()
        if self.manager is not None:
            self.manager.shutdown()

    @property
    def manager(self) -> SyncManager:
        if self._manager is None:
            self._manager = multiprocessing.Manager()
        return self._manager

    def create_queue(self) -> Queue:
        return self.manager.Queue()

    def initialize(self):
        self._process_queue = self.create_queue()
        self._process_result_queue = self.create_queue()
        self.process_channel = ChannelQueue(channel_queue=self._process_queue)
        self.process_result_channel = ChannelQueue(channel_queue=self._process_result_queue)
        self._threads: List[ThreadInfo] = []
        self._thread_tasks: List[ThreadTask] = []


    def start(self, target_method, target_kwargs, process_count=1):
        self.initialize()
        # Initiate the worker processes
        for i in range(process_count):
            thread_id = i + 1
            worker = ParallelWorker(
                thread_id=thread_id,
                operation_integration=self.operation_integration,
                channel=self.channel,
                process_channel=self.process_channel,
                process_result_channel=self.process_result_channel,
                target_method=target_method,
                target_kwargs=target_kwargs,
            )

            # Add new process to the list of processes
            thread_info = ThreadInfo(
                Id=thread_id,
                Thread=worker
            )
            self._threads.append(thread_info)
            process_task = ThreadTask(
                Id=thread_id,
                IsFinished=False
            )
            self.process_channel.put(process_task)
            self._thread_tasks.append(process_task)
            # Start the process
        for t in self._threads:
            t.Thread.start()

    def get_results(self) -> List[ThreadTask]:
        self.__check_threads()
        return self._thread_tasks

    def __check_threads(self):
        # Read calculation results
        while True:
            # Read result
            thread_task: ThreadTask = self.process_result_channel.get()
            # Have a look at the results
            if thread_task.IsFinished:
                # Process has finished
                if thread_task.State == 4:
                    self.__finish_all_threads()
                    for task in self._thread_tasks:
                        if task.Id == thread_task.Id:
                            task.Result = thread_task.Result
                            task.State = thread_task.State
                            task.Message = thread_task.Message
                            task.IsFinished = thread_task.IsFinished
                            task.Exception = thread_task.Exception
                            task.Traceback = thread_task.Traceback
                            task.Data = thread_task.Data
                    break
                for t in self._threads:
                    if t.Id == thread_task.Id:
                        t.IsFinished = True
                for task in self._thread_tasks:
                    if task.Id == thread_task.Id:
                        task.Result = thread_task.Result
                        task.State = thread_task.State
                        task.Message = thread_task.Message
                        task.IsFinished = thread_task.IsFinished
                        task.Exception = thread_task.Exception
                        task.Traceback = thread_task.Traceback
                        task.Data = thread_task.Data
                self.__check_unfinished_threads()
                if self.__check_all_threads_finished():
                    break

    def __finish_all_threads(self):
        if self._threads is not None:
            for t in self._threads:

                if t is not None and t.Thread.is_alive() and not t.Thread.stopped():
                    t.Thread.stop()
                if t.Thread.is_alive():
                    process_task = ThreadTask(
                        Id=t.Id,
                        IsFinished=False
                    )
                    self.process_channel.put(process_task)

                    self.publisher.publish(
                        message=TaskMessage(
                            event=EVENT_LOG,
                            kwargs={
                                'data': self.operation_integration,
                                'message': f"Process will terminate. WorkerId:{t.Id}"
                            }
                        )
                    )
                    t.IsFinished = True
                    t.Thread.stop()

    def __check_unfinished_threads(self):

        for t in self._threads:
            if not t.IsFinished and not t.Thread.is_alive():
                self.publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': self.operation_integration,
                            'message': f"Unfinished process found. WorkerId:{t.Id}"
                        }
                    )
                )
                t.IsFinished = True

    def __check_all_threads_finished(self):
        check_finish = True
        unfinished_thread_list = []
        for t in self._threads:
            if not t.IsFinished:
                check_finish = False
                unfinished_thread_list.append(str(t.Id))
        thread_join = ",".join(unfinished_thread_list)

        self.publisher.publish(
            message=TaskMessage(
                event=EVENT_LOG,
                kwargs={
                    'data': self.operation_integration,
                    'message': f"All processes are expected to end . WorkerId:{thread_join}"
                }
            )
        )
        return check_finish



from abc import abstractmethod

from ...base import Initializer
from ....operation.domain import OperationBase


class OperationExecutionInitializer(Initializer):
    @abstractmethod
    def initialize(self, operation: OperationBase) -> OperationBase:
        pass



# This decorator will check unexpected database error for thread operations
from ..repository import RepositoryProvider
from ...dependency.container import DependencyContainer


def transactionhandler(func):
    def inner(*args, **kwargs):
        repository_provider = DependencyContainer.Instance.get(RepositoryProvider)
        try:
            result = func(*args, **kwargs)
            repository_provider.commit()
            return result
        except Exception as ex:
            repository_provider.rollback()
            raise
        finally:
            repository_provider.close()

    return inner



import os

from injector import inject

from pdip.configuration.models.application import ApplicationConfig
from pdip.dependency import IScoped
from pdip.integrator.connection.domain.enums import ConnectorTypes
from pdip.logging.loggers.sql import SqlLogger
from .file_connector import FileConnector
from .file_context import FileContext
from ..connectors.csv import CsvConnector


class FileProvider(IScoped):
    @inject
    def __init__(self,
                 logger: SqlLogger,
                 application_config: ApplicationConfig
                 ):
        self.application_config = application_config
        self.logger = logger

    def get_context(self, connection, connection_basic_authentication, connection_server) -> FileContext:
        """
        Creating Connection
        """
        connector: FileConnector = None
        if connection.File.ConnectorType.Name == ConnectorTypes.CSV.name:
            host = connection_server.Host
            port = connection_server.Port
            if host is None or host == '':
                host = os.path.join(self.application_config.root_directory, "files")
            if connection.File.ConnectorType.Name == ConnectorTypes.CSV.name:
                connector = CsvConnector(host=host)
        if connector is not None:
            file_context: FileContext = FileContext(connector=connector)
            return file_context
        else:
            raise Exception(f"{connection.File.ConnectorType.Name} connector type not supported")

    def get_file_context_with_host(self, host: str) -> FileContext:
        """
        Creating Connection
        """
        if host is None or host == '':
            host = os.path.join(self.application_config.root_directory, "files")
        connector = CsvConnector(folder=host)
        file_context: FileContext = FileContext(connector=connector)
        return file_context



from injector import inject

from ...api.request_parameter import PagingParameter
from ...dependency import IScoped


class PagingSpecification(IScoped):
    @inject
    def __init__(self):
        self.default_page_size = 10
        self.default_page_number = 1
        self.default_offset = 0
        self.min_page_size = 5
        self.max_page_size = 200
        pass

    def specify(self, paging_parameter: PagingParameter):
        if not hasattr(paging_parameter, 'PageSize') and hasattr(paging_parameter, 'PageNumber'):
            return None, None
        if paging_parameter.PageSize is None or paging_parameter.PageSize < self.min_page_size or paging_parameter.PageSize > self.max_page_size:
            page_size = self.default_page_size
        else:
            page_size = paging_parameter.PageSize
        if paging_parameter.PageNumber is None or paging_parameter.PageNumber < self.default_page_number:
            page_number = self.default_page_number
        else:
            page_number = paging_parameter.PageNumber
        offset = (page_number - 1) * page_size
        if offset is None or offset < self.default_offset:
            offset = self.default_offset
        return page_size, offset



import logging
import traceback
from logging import handlers

from injector import inject

from ..base import ILogger
from ....dependency import ISingleton
from ....utils import Utils


class FileLogger(ISingleton, ILogger):
    @inject
    def __init__(self, log_level=logging.DEBUG):
        self.log_level = log_level
        self.logger = logging.getLogger('pdip.file_logger')
        self.file_handler = None
        self.log_init()

    def __del__(self):
        self.logger.removeHandler(self.file_handler)
        del self.file_handler

    def log_init(self):
        """
        initialization of log file.
        """
        if not self.logger.handlers:
            self.logger.setLevel(self.log_level)
            process_info = Utils.get_process_info()
            format = logging.Formatter(
                f"%(asctime)s:%(name)s:%(levelname)s: - {process_info} - %(message)s")

            self.file_handler = handlers.RotatingFileHandler(
                'pdip.log', maxBytes=(1048576 * 5), backupCount=7)
            self.file_handler.setFormatter(format)
            self.logger.addHandler(self.file_handler)

    def log(self, level, message):
        self.logger.log(level, message)

    def exception(self, exception: Exception, message: str = None, job_id=None):
        exc = traceback.format_exc() + '\n' + str(exception)
        if message is not None:
            message += f"Error: {exc}"
        else:
            message = f"Error: {exc}"
        self.logger.error(message)

    def critical(self, message):
        self.logger.critical(message)

    def fatal(self, message):
        self.logger.fatal(message)

    def error(self, message):
        self.logger.error(message)

    def warning(self, message):
        self.logger.warning(message)

    def info(self, message):
        self.logger.info(message)

    def debug(self, message):
        self.logger.debug(message)



from .incompatible_adapter_exception import IncompatibleAdapterException
from .not_supported_feature_exception import NotSupportedFeatureException
from .operational_exception import OperationalException
from .required_class_exception import RequiredClassException



import json
import typing
from datetime import datetime
import uuid
from flask import request
from flask_restx import Api, fields, inputs
from flask_restx.reqparse import RequestParser, Argument
from injector import inject

from ...api.converter import RequestConverter
from ...api.request_parameter import OrderByParameter
from ...api.request_parameter import PagingParameter
from ...utils import TypeChecker

T = typing.TypeVar('T')


class NullableString(fields.String):
    __schema_type__ = ['string', 'null']
    __schema_example__ = 'nullable string'


class EndpointWrapper:
    @inject
    def __init__(self, api: Api) -> None:
        self.api = api
        self.type_checker = TypeChecker()

    @property
    def BaseModel(cls):
        return cls.api.model('BaseModel', {
            'IsSuccess': fields.Boolean(description='Is Success', default=True),
            'Message': fields.String(description='Message', default="Operation Completed"),
            'Result': fields.Raw(description='Service result values'),
        })

    @staticmethod
    def date_converter(o):
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, uuid.UUID):
            return str(o)

    @staticmethod
    def get_response(result=None, message=None):
        return {'Result': result, 'Message': message}

    def field_resolver(self, value, key):

        if value == int:
            specified_value = fields.Integer(description=f'{key}')
        elif value == str:
            specified_value = NullableString(description=f'{key}')
        elif value == bool:
            specified_value = fields.Boolean(description=f'{key}')
        elif value == datetime:
            specified_value = fields.DateTime(description=f'{key}',
                                              example=datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ'))
        elif value == float:
            specified_value = fields.Float(description=f'{key}')
        elif value == any:
            specified_value = fields.Raw(description=f'{key}')
        return specified_value

    def annotation_resolver(self, annotations):
        definition = {}
        for key in annotations:
            value = annotations[key]
            specified_value = None
            if self.type_checker.is_primitive(value) or value == datetime:
                specified_value = self.field_resolver(value, key)
            elif self.type_checker.is_generic(value):
                if value.__args__[0] is any:
                    specified_value = fields.List(fields.Raw(), description=f'')
                else:
                    instance = value.__args__[0]()
                    nested_annotations = self.get_annotations(instance)
                    if nested_annotations is not None:
                        nested_model_definition = self.annotation_resolver(nested_annotations)
                        nested_model = self.api.model(value.__args__[0].__name__, nested_model_definition)
                        specified_value = fields.List(fields.Nested(nested_model), description=f'')
                    elif self.type_checker.is_primitive(instance.__class__):
                        field = self.field_resolver(instance.__class__, key)
                        specified_value = fields.List(field, description=f'')

            elif self.type_checker.is_base_generic(value):
                # TODO:Base generic class
                print('value type should be a structure of', value.__args__[0])
            elif self.type_checker.is_class(value):
                instance = value()
                nested_annotations = self.get_annotations(instance)
                if nested_annotations is not None:
                    nested_model_definition = self.annotation_resolver(nested_annotations)
                    nested_model = self.api.model(value.__name__, nested_model_definition)
                    specified_value = fields.Nested(nested_model, description=f'')
            else:
                print('Type not know', value)
            if specified_value is not None:
                definition[key] = specified_value
        return definition

    def request_model(self, model_type: typing.Type[T]) -> RequestParser:
        instance = model_type()
        annotations = self.get_annotations(instance)
        if annotations is not None:
            model_definition = self.annotation_resolver(annotations)
            model = self.api.model(model_type.__name__, model_definition)
            return model

    def response_model(self, model_type: typing.Type[T]) -> RequestParser:
        if self.type_checker.is_class(model_type):
            instance = model_type()
            annotations = self.get_annotations(instance)
            if annotations is not None:
                model_definition = self.annotation_resolver(annotations)
                model = self.api.model(model_type.__name__, model_definition)
                success_model = self.api.model(model_type.__name__ + 'Base', {
                    'IsSuccess': fields.Boolean(description='Is Success', default=True),
                    'Message': fields.String(description='Message', default=None),
                    'Result': fields.Nested(model, description='Result'),
                })
                return success_model
        else:
            # success_model = self.api.model(model_type.__name__ + 'Base', {
            #     'IsSuccess': fields.Boolean(description='Is Success', default=True),
            #     'Message': fields.String(description='Message', default=None),
            #     'Result': fields.Raw(description=f'Result'),
            # })
            return self.BaseModel

    def create_parser(self, name, input_type: typing.Type[T]) -> RequestParser:
        parser: RequestParser = self.api.parser()

        if TypeChecker().is_class(input_type) == True:
            parser.add_argument(self.create_argument(name=name, type=input_type, location='form', help=name))
        else:
            parser.add_argument(self.create_argument(name=name, type=input_type, location='args', help=name))
        return parser

    def get_request_from_parser_for_primitive(self, name, input_type: typing.Type[T]) -> T:
        parser = self.create_parser(name=name, input_type=input_type)
        data = parser.parse_args(request)
        return data[name]

    def create_argument(self, name, type, location, help) -> Argument:
        specified_type = type
        if TypeChecker().is_generic(type) == True:
            specified_type = type.__args__[0]

        if specified_type == bool:
            specified_type = inputs.boolean
        argument = Argument(name, type=specified_type, location=location, help=help)
        return argument

    def get_annotations(self, obj):
        if hasattr(obj, '__annotations__'):
            annotations = obj.__annotations__
            return annotations
        else:
            return None

    def request_parser(self, parser_type: typing.Type[T]) -> RequestParser:
        parser: RequestParser = self.api.parser()
        instance = parser_type()
        if isinstance(instance, PagingParameter):
            parser.add_argument('PageNumber', type=int, location='args', help='Page Number')
            parser.add_argument('PageSize', type=int, location='args', help='Page Size')
        if isinstance(instance, OrderByParameter):
            parser.add_argument('OrderBy', type=str, location='args', help='Order By')
            parser.add_argument('Order', type=str, location='args', help='Order')
        annotations = self.get_annotations(instance)
        if annotations is not None:
            for name in annotations:
                value = annotations[name]
                parser.add_argument(self.create_argument(name=name, type=value, location='args', help=name))
        return parser

    def get_request_from_parser(self, parser_type: typing.Type[T]) -> T:
        request_converter = RequestConverter()
        request_converter.register(parser_type)
        data = self.request_parser(parser_type).parse_args(request)
        json_data = json.dumps(data)
        req: T = request_converter.FromJSON(json_data)
        # req: T = json.loads(json_data, object_hook=lambda d: parser_type(**d))
        return req

    def get_request_from_body(self, parser_type: typing.Type[T]) -> T:
        request_converter = RequestConverter()
        request_converter.register(parser_type)
        data = self.api.payload
        json_data = json.dumps(data)
        req: T = request_converter.FromJSON(json_data)
        # req: T = json.loads(json_data, object_hook=lambda d: parser_type(**d))
        return req



import os
import sys

from .endpoint_base import endpoint
from ...configuration.models.application import ApplicationConfig
from ...utils import Utils


class Controller:
    def __init__(self,
                 cls,
                 api,
                 application_config: ApplicationConfig,
                 namespace=None,
                 namespace_name=None,
                 route=None
                 ):
        self.route = route
        self.cls = cls
        self.namespace = namespace
        self.namespace_name = namespace_name
        self.application_config = application_config
        self.api = api
        self.white_list = ['get', 'post', 'put', 'delete']

    def create_route(self):

        controller_namespace = self.namespace
        controller_namespace_name = None
        if self.namespace is not None:
            controller_namespace = self.namespace
            controller_namespace_name = self.namespace.name
        elif self.namespace_name is not None and self.namespace_name != '':
            controller_namespace, controller_namespace_name = self.find_namespace(name=self.namespace_name)
        else:
            controller_namespace, controller_namespace_name = self.find_namespace(name=None)
        controller_route = self.route
        if controller_route is None:
            controller_route = self.find_route(controller_namespace_name)

        decorator = endpoint(namespace=controller_namespace, api=self.api)
        for attr in self.cls.__dict__:
            if callable(getattr(self.cls, attr)) and attr in self.white_list:
                func_attr = getattr(self.cls, attr)
                if hasattr(func_attr, 'decorated') and func_attr.decorated:
                    continue
                setattr(self.cls, attr, decorator(func_attr))

        return controller_namespace.route(controller_route)(self.cls)

    def find_namespace(self, name):
        if name is not None:
            namespace_name = name
        else:
            namespace_name = self.get_namespace_name()
        founded_namespace = None
        for api_namespace in self.api.namespaces:
            if api_namespace.name == namespace_name:
                founded_namespace = api_namespace
                break
        else:
            founded_namespace = self.api.namespace(namespace_name,
                                                   description=f'{namespace_name} endpoints',
                                                   path=f'/api/{namespace_name}')
        return founded_namespace, namespace_name

    def get_namespace_name(self):
        excluded_namespace_names = ['controllers', 'Controllers']
        path_of_root = os.path.join(self.application_config.root_directory, '')
        module_name = self.cls.__module__
        module_path = sys.modules[module_name].__file__
        namespace_folder = module_path.lower().replace(path_of_root.lower(), '')
        split_namespace = Utils.path_split(namespace_folder)
        for excluded_namespace_name in excluded_namespace_names:
            if excluded_namespace_name in split_namespace:
                split_namespace.remove(excluded_namespace_name)
        if len(split_namespace) > 1:
            name = split_namespace[len(split_namespace) - 2].title()
        else:
            name = self.cls.__name__.replace(
                'Controller', '').replace('Resource', '')
        return name

    def find_route(self, namespace_name):
        route_path = self.cls.__name__.replace('Controller', '').replace('Resource', '').replace(namespace_name, '')
        return f'/{route_path}' if route_path is not None and route_path != '' else ''



from func_timeout import func_set_timeout
from injector import inject

from ...base import IntegrationSourceToTargetExecuteStrategy
from .......connection.factories import ConnectionSourceAdapterFactory, ConnectionTargetAdapterFactory
from .......domain.enums.events import EVENT_LOG
from .......operation.domain.operation import OperationIntegrationBase
from .......pubsub.base import ChannelQueue
from .......pubsub.domain import TaskMessage
from .......pubsub.publisher import Publisher
from ........dependency import IScoped


class SingleProcessIntegrationExecute(IntegrationSourceToTargetExecuteStrategy, IScoped):
    @inject
    def __init__(self,
                 connection_source_adapter_factory: ConnectionSourceAdapterFactory,
                 connection_target_adapter_factory: ConnectionTargetAdapterFactory
                 ):
        self.connection_source_adapter_factory = connection_source_adapter_factory
        self.connection_target_adapter_factory = connection_target_adapter_factory

    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ) -> int:
        publisher = Publisher(channel=channel)
        try:
            source_adapter = self.connection_source_adapter_factory.get_adapter(
                connection_type=operation_integration.Integration.SourceConnections.ConnectionType
            )
            target_adapter = self.connection_target_adapter_factory.get_adapter(
                connection_type=operation_integration.Integration.TargetConnections.ConnectionType
            )
            integration = operation_integration.Integration
            limit = operation_integration.Limit
            iterator = source_adapter.get_iterator(
                integration=integration,
                limit=limit
            )

            task_id = 0
            start = 0
            data_count = 0
            end = limit
            for results in iterator:
                task_id += 1
                data_length = len(results)
                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"0 - data :{task_id}-{start}-{end} readed from db"
                        }
                    )
                )
                self.write_target_data(
                    target_adapter=target_adapter,
                    integration=operation_integration.Integration,
                    source_data=results
                )
                publisher.publish(
                    message=TaskMessage(
                        event=EVENT_LOG,
                        kwargs={
                            'data': operation_integration,
                            'message': f"0 - data :{task_id}-{start}-{end} process finished task. "
                        }
                    )
                )
                data_count += data_length
                end += limit
                start += limit

            return data_count
        except Exception as ex:
            publisher.publish(
                message=TaskMessage(
                    event=EVENT_LOG,
                    kwargs={
                        'data': operation_integration,
                        'message': f"Integration getting error. ",
                        'exception': ex
                    }
                )
            )
            raise

    @func_set_timeout(1800)
    def write_target_data(
            self,
            target_adapter,
            integration,
            source_data
    ):
        target_adapter.write_data(integration=integration,
                                  source_data=source_data)



import json
import traceback

from injector import inject

from ...data.repository import RepositoryProvider
from ...dependency import ISingleton
from ...dependency.provider import ServiceProvider
from ...logging.loggers.sql import SqlLogger


class ErrorHandlers(ISingleton):
    @inject
    def __init__(
            self,
            logger: SqlLogger,
            service_provider: ServiceProvider
    ):
        self.service_provider = service_provider
        self.logger = logger
        self.separator = '|'
        self.default_content_type = "application/json"
        self.mime_type_string = "mimetype"

    def handle_http_exception(self, exception):
        """Return JSON instead of HTML for HTTP errors."""
        # start with the correct headers and status code from the error
        response = exception.get_response()
        # replace the body with JSON
        response.data = json.dumps({
            "Result": "",
            "IsSuccess": "false",
            "Code": exception.code,
            "Name": exception.name,
            "Message": exception.description,
        })
        message = "empty"
        if exception is not None and exception.description is not None:
            message = exception.description
        self.logger.error(f'code:{exception.code} - name:{exception.name} - Message:{message}')
        response.content_type = self.default_content_type
        return response

    def handle_exception(self, exception):
        """Return JSON instead of HTML for HTTP errors."""
        self.service_provider.get(RepositoryProvider).rollback()
        # start with the correct headers and status code from the error
        exception_traceback = traceback.format_exc()
        output = self.separator.join(exception.args)
        # replace the body with JSON
        # response = json.dumps({
        #     "result": "",
        #     "isSuccess": "false",
        #     "message": f"ConnectionServer exception occurred. Exception Message:{output}",
        # })
        output_message = "empty"
        if output is not None and output != "":
            output_message = output
        trace_message = "empty"
        if exception_traceback is not None and exception_traceback != "":
            trace_message = exception_traceback
        self.logger.error(f'Messsage:{output_message} - Trace:{trace_message}')
        return {
                   "Result": "",
                   "IsSuccess": "false",
                   "Message": f"ConnectionServer exception occurred. Exception Message:{output}",
               }, 500, {self.mime_type_string: self.default_content_type}

    def handle_operational_exception(self, exception):
        separator = '|'
        default_content_type = "application/json"
        mime_type_string = "mimetype"
        """Return JSON instead of HTML for HTTP errors."""
        self.service_provider.get(RepositoryProvider).rollback()
        # start with the correct headers and status code from the error
        exception_traceback = traceback.format_exc()
        output = separator.join(exception.args)
        output_message = "empty"
        if output is not None and output != "":
            output_message = output
        trace_message = "empty"
        if exception_traceback is not None and exception_traceback != "":
            trace_message = exception_traceback
        self.logger.error(
            f'Operational Exception Messsage:{output_message} - Trace:{trace_message}')
        return {
                   "Result": "",
                   "IsSuccess": "false",
                   "Message": output
               }, 400, {mime_type_string: default_content_type}



from injector import inject

from pdip.dependency import IScoped
from pdip.integrator.connection.factories import ConnectionSourceAdapterFactory
from pdip.integrator.domain.enums.events import EVENT_LOG
from pdip.integrator.integration.domain.base import IntegrationBase
from pdip.integrator.operation.domain import OperationIntegrationBase
from pdip.integrator.pubsub.base import ChannelQueue
from pdip.integrator.pubsub.domain import TaskMessage
from pdip.integrator.pubsub.publisher import Publisher
from ...base import IntegrationAdapter


class SourceIntegration(IntegrationAdapter, IScoped):
    @inject
    def __init__(self,
                 connection_source_adapter_factory: ConnectionSourceAdapterFactory
                 ):
        self.connection_source_adapter_factory = connection_source_adapter_factory

    def execute(
            self,
            operation_integration: OperationIntegrationBase,
            channel: ChannelQueue
    ) -> int:
        publisher = Publisher(channel=channel)
        source_adapter = self.connection_source_adapter_factory.get_adapter(
            connection_type=operation_integration.Integration.SourceConnections.ConnectionType)
        order = operation_integration.Order
        row_count = source_adapter.get_source_data_count(integration=operation_integration.Integration)

        publisher.publish(message=TaskMessage(event=EVENT_LOG,
                                              kwargs={
                                                  'data': operation_integration,
                                                  'message': f"{order} - source has {row_count}"
                                              }))
        return row_count

    def get_start_message(self, integration: IntegrationBase):
        message = f"Integration execute started."
        if integration.SourceConnections.Sql is not None:
            message = f"{integration.SourceConnections.Sql.Schema}.{integration.SourceConnections.Sql.ObjectName} integration execute started."
        elif integration.SourceConnections.BigData is not None:
            message = f"{integration.SourceConnections.BigData.Schema}.{integration.SourceConnections.BigData.ObjectName} integration execute started."
        elif integration.SourceConnections.WebService is not None:
            message = f"{integration.SourceConnections.WebService.Method} integration execute started."
        elif integration.SourceConnections.File is not None:
            message = f"{integration.SourceConnections.File.Folder}\\{integration.SourceConnections.File.FileName} integration execute started."
        elif integration.SourceConnections.Queue is not None:
            message = f"{integration.SourceConnections.Queue.TopicName} integration execute started."
        return message

    def get_finish_message(self, integration: IntegrationBase, data_count: int):
        message = f"Integration execute finished"
        if integration.SourceConnections.Sql is not None:
            message = f"{integration.SourceConnections.Sql.Schema}.{integration.SourceConnections.Sql.ObjectName} integration execute finished."
        elif integration.SourceConnections.BigData is not None:
            message = f"{integration.SourceConnections.BigData.Schema}.{integration.SourceConnections.BigData.ObjectName} integration execute finished."
        elif integration.SourceConnections.BigData is not None:
            message = f"{integration.SourceConnections.WebService.Method} integration execute finished."
        elif integration.SourceConnections.File is not None:
            message = f"{integration.SourceConnections.File.Folder}\\{integration.SourceConnections.File.FileName} integration execute finished."
        elif integration.SourceConnections.Queue is not None:
            message = f"{integration.SourceConnections.Queue.TopicName} integration execute finished."

        return message

    def get_error_message(self, integration: IntegrationBase):
        message = f"Integration execute getting error."
        if integration.SourceConnections.Sql is not None:
            message = f"{integration.SourceConnections.Sql.Schema}.{integration.SourceConnections.Sql.ObjectName} integration execute getting error."
        elif integration.SourceConnections.BigData is not None:
            message = f"{integration.SourceConnections.BigData.Schema}.{integration.SourceConnections.BigData.ObjectName} integration execute getting error."
        if integration.SourceConnections.WebService is not None:
            message = f"{integration.SourceConnections.WebService.Method} integration execute getting error."
        elif integration.SourceConnections.File is not None:
            message = f"{integration.SourceConnections.File.Folder}\\{integration.SourceConnections.File.FileName} integration execute getting error."
        elif integration.SourceConnections.Queue is not None:
            message = f"{integration.SourceConnections.Queue.TopicName} integration execute getting error."
        return message



from .execution import EventBase
from .execution import ExecutionBase
from .execution import ExecutionOperationBase
from .execution import ExecutionOperationEvent
from .execution import ExecutionOperationIntegrationBase
from .execution import ExecutionOperationIntegrationEvent



from .base_config import BaseConfig
from ..api.api_config import ApiConfig
from ..application.application_config import ApplicationConfig
from ..aps.aps_config import ApsConfig
from ..database.database_config import DatabaseConfig



import builtins
import inspect
import typing
from abc import ABC


class ITypeChecker(ABC):
    def is_class(self, obj):
        if inspect.isclass(obj) and not self.is_primitive(obj):
            return True
        return False

    def is_primitive(self, obj):
        builtins_list = list(
            filter(lambda x: not x.startswith('_'), dir(builtins)))

        return hasattr(obj, '__name__') and obj.__name__ in builtins_list

    def is_generic(self, class_type):
        pass

    def is_base_generic(self, class_type):
        pass

class TypeChecker(ITypeChecker):
    def is_generic(self, class_type):
        return self._is_generic(class_type)

    def is_base_generic(self, class_type):
        return self._is_base_generic(class_type)

    def _is_generic(self, cls):
        if isinstance(cls, typing._GenericAlias):
            return True
        if isinstance(cls, typing._SpecialForm):
            return cls not in {typing.Any}
        return False

    def _is_base_generic(self, class_type):
        if isinstance(class_type, typing._GenericAlias):
            if class_type.__origin__ in {typing.Generic, typing._Protocol}:
                return False
            if isinstance(class_type, typing._VariadicGenericAlias):
                return True
            return len(class_type.__parameters__) > 0
        if isinstance(class_type, typing._SpecialForm):
            return class_type._name in {'ClassVar', 'Union', 'Optional'}
        return False



from injector import inject

from .big_data_context import BigDataContext
from .big_data_policy import BigDataPolicy
from ....domain.authentication.basic import ConnectionBasicAuthentication
from ....domain.authentication.kerberos import KerberosAuthentication
from ....domain.authentication.mechanism import MechanismTypes
from ....domain.enums import ConnectorTypes, ConnectionTypes
from ....domain.server.base import ConnectionServer
from ....domain.types.bigdata.configuration.base import BigDataConnectionConfiguration
from ......dependency import IScoped


class BigDataProvider(IScoped):
    @inject
    def __init__(self):
        pass

    def __initialize_context(self, config: BigDataConnectionConfiguration):
        policy = BigDataPolicy(config=config)
        context = BigDataContext(policy=policy)
        return context

    def get_context_by_config(self, config: BigDataConnectionConfiguration) -> BigDataContext:
        return self.__initialize_context(config=config)

    def get_context(
            self,
            connector_type: ConnectorTypes, mechanism_type: MechanismTypes,
            host: str, port: int,
            user: str, password: str,
            database: str,
            ssl: bool, use_only_sspi: bool) -> BigDataContext:
        """
        Creating Context
        """
        if connector_type == connector_type.Impala:
            config = BigDataConnectionConfiguration(
                ConnectionType=ConnectionTypes.BigData,
                ConnectorType=ConnectorTypes.Impala,
                AuthenticationMechanismType=mechanism_type,
                Server=ConnectionServer(Host=host, Port=port),
                BasicAuthentication=ConnectionBasicAuthentication(User=user,
                                                                  Password=password),
                KerberosAuthentication=KerberosAuthentication(Principal=user, Password=password),
                Database=database,
                Ssl=ssl,
                UseOnlySspi=use_only_sspi
            )
        else:
            raise Exception(f"{connector_type.name} connector type not supported")

        return self.__initialize_context(config=config)



from typing import Generic, TypeVar

from .command_query_handler_base import CommandQueryHandlerBase
from .icommand import ICommand

CH = TypeVar('CH', covariant=True, bound=ICommand)


class ICommandHandler(Generic[CH], CommandQueryHandlerBase[CH]):
    def __init__(self) -> CH:
        pass

    def handle(self, query: CH) -> CH:
        pass



from dataclasses import dataclass

from .cls_to_dict import cls_to_dict
from pdip.json.base.json_convert import JsonConvert


def requestclass(_cls=None):
    def wrap(cls):
        return (JsonConvert.register)(cls_to_dict(cls=(dataclass)(cls)))

    return wrap(_cls)
